import struct

from functools import reduce
from operator import mul
from typing import List, Tuple

from .codecs.string import encode_str
from .codecs.utils import InputOrOutput
from .codecs.lists import ListElement

_DatatypeToCtype = {
    "BOOL": "?",
    "UINT8": "B",
    "UINT16": "H",
    "UINT32": "I",
    "UINT64": "L",
    "INT8": "b",
    "INT16": "h",
    "INT32": "i",
    "INT64": "l",
    "FP16": "e",
    "FP32": "f",
    "FP64": "d",
}

_SizeFormat = "<I"


def _tensor_length(shape: List[int]) -> int:
    return reduce(mul, shape, 1)


def _tensor_format(elem: InputOrOutput) -> str:
    size = _tensor_length(elem.shape)
    ctype = _DatatypeToCtype[elem.datatype]
    return f"{size}{ctype}"


def _unpack_bytes(raw: bytes) -> List[bytes]:
    """
    From Triton's implementation:
        https://github.com/triton-inference-server/client/blob/6cc412c50ca4282cec6e9f62b3c2781be433dcc6/src/python/library/tritonclient/utils/__init__.py#L246-L273
    """
    offset = 0
    length = len(raw)
    elems = []
    while offset < length:
        [size] = struct.unpack_from(_SizeFormat, raw, offset)
        offset += 4
        [elem] = struct.unpack_from(f"<{size}s", raw, offset)
        offset += size

        elems.append(elem)

    return elems


def _pack_bytes(unpacked: List[ListElement]) -> bytes:
    packed = []
    for elem in unpacked:
        as_bytes = _ensure_bytes(elem)
        size = len(as_bytes)

        packed_size = struct.pack(_SizeFormat, size)
        packed.append(packed_size)

        packed_elem = struct.pack(f"<{size}s", as_bytes)
        packed.append(packed_elem)

    return b"".join(packed)


def _ensure_bytes(elem: ListElement) -> bytes:
    if isinstance(elem, str):
        return encode_str(elem)

    return elem


def _unpack_tensor(elem: InputOrOutput, raw: bytes) -> list:
    tensor_format = _tensor_format(elem)
    return list(struct.unpack(tensor_format, raw))


def _pack_tensor(elem: InputOrOutput) -> bytes:
    tensor_format = _tensor_format(elem)
    return struct.pack(tensor_format, *elem.data)


def unpack(elem: InputOrOutput, raw: bytes) -> list:
    if elem.datatype == "BYTES":
        return _unpack_bytes(raw)

    return _unpack_tensor(elem, raw)


def pack(elem: InputOrOutput) -> bytes:
    if elem.datatype == "BYTES":
        return _pack_bytes(elem.data)  # type: ignore

    return _pack_tensor(elem)


def inject_raw(
    elems: List[InputOrOutput], raw_contents: List[bytes]
) -> List[InputOrOutput]:
    raw_idx = 0
    for elem in elems:
        if not elem.data:
            # Only unpack raw entry, if input / output is empty
            # This is to allow for mixed reqs / resp, where some entries have
            # data, some entries are raw
            raw = raw_contents[raw_idx]
            elem.data = unpack(elem, raw)  # type: ignore
            raw_idx += 1

    return elems


def extract_raw(elems: List[InputOrOutput]) -> Tuple[List[InputOrOutput], List[bytes]]:
    raw_contents = []
    for elem in elems:
        raw = pack(elem)
        raw_contents.append(raw)
        elem.data = []  # type: ignore

    return elems, raw_contents



from typing import Any, Union, List
from datetime import datetime
from functools import partial

from ..types import RequestInput, ResponseOutput
from .lists import is_list_of, as_list, ListElement
from .base import InputCodec, register_input_codec

_Datetime = Union[str, datetime]
_DatetimeStrCodec = "ascii"


def _to_iso(elem: _Datetime) -> str:
    if isinstance(elem, str):
        # Assume date has already been encoded upstream
        return elem

    return elem.isoformat()


def _encode_datetime(elem: _Datetime, use_bytes: bool) -> Union[bytes, str]:
    iso_date = _to_iso(elem)
    if not use_bytes:
        return iso_date

    return iso_date.encode(_DatetimeStrCodec)


def _ensure_str(elem: ListElement) -> str:
    if isinstance(elem, str):
        return elem

    return elem.decode(_DatetimeStrCodec)


def _decode_datetime(elem: ListElement) -> datetime:
    iso_format = _ensure_str(elem)
    return datetime.fromisoformat(iso_format)


@register_input_codec
class DatetimeCodec(InputCodec):
    """
    Codec that convers to / from a datetime input.
    """

    ContentType = "datetime"
    TypeHint = List[_Datetime]

    @classmethod
    def can_encode(cls, payload: Any) -> bool:
        return is_list_of(payload, datetime)

    @classmethod
    def encode_output(
        cls, name: str, payload: List[_Datetime], use_bytes: bool = True, **kwargs
    ) -> ResponseOutput:
        packed = map(partial(_encode_datetime, use_bytes=use_bytes), payload)
        shape = [len(payload), 1]
        return ResponseOutput(
            name=name,
            datatype="BYTES",
            shape=shape,
            data=list(packed),
        )

    @classmethod
    def decode_output(cls, response_output: ResponseOutput) -> List[datetime]:
        packed = response_output.data.__root__

        return list(map(_decode_datetime, as_list(packed)))

    @classmethod
    def encode_input(
        cls, name: str, payload: List[_Datetime], use_bytes: bool = True, **kwargs
    ) -> RequestInput:
        output = cls.encode_output(name, payload, use_bytes=use_bytes)
        return RequestInput(
            name=output.name,
            datatype=output.datatype,
            shape=output.shape,
            data=output.data,
        )

    @classmethod
    def decode_input(cls, request_input: RequestInput) -> List[datetime]:
        packed = request_input.data.__root__

        return list(map(_decode_datetime, as_list(packed)))



from typing import Optional, Tuple

from ..utils import insert_headers, extract_headers
from ..types import InferenceRequest
from ..handlers import DataPlane

from .message import KafkaMessage
from .errors import InvalidMessageHeaders

MLSERVER_MODEL_NAME_HEADER = "mlserver-model"
MLSERVER_MODEL_VERSION_HEADER = "mlserver-version"


class KafkaHandlers:
    def __init__(self, data_plane: DataPlane):
        self._data_plane = data_plane

    async def infer(self, request: KafkaMessage) -> KafkaMessage:
        inference_request = InferenceRequest(**request.value)

        # Kafka KEY takes precedence over body ID
        if request.key:
            inference_request.id = request.key

        insert_headers(inference_request, request.headers)

        model_name, model_version = self._get_model_details(request)
        inference_response = await self._data_plane.infer(
            inference_request, model_name, model_version
        )

        response_headers = extract_headers(inference_response) or {}
        return KafkaMessage.from_types(
            inference_response.id, inference_response, response_headers
        )

    def _get_model_details(self, request: KafkaMessage) -> Tuple[str, Optional[str]]:
        headers = request.headers

        # TODO: Update header with consistency with other headeres
        if MLSERVER_MODEL_NAME_HEADER not in headers:
            raise InvalidMessageHeaders(MLSERVER_MODEL_NAME_HEADER)

        model_name = headers[MLSERVER_MODEL_NAME_HEADER]
        model_version = headers.get(MLSERVER_MODEL_VERSION_HEADER, None)

        return model_name, model_version



from ..settings import Settings
from ..registry import MultiModelRegistry
from ..types import (
    MetadataModelResponse,
    MetadataServerResponse,
    InferenceRequest,
    InferenceResponse,
)
from ..middleware import InferenceMiddlewares
from ..cloudevents import CloudEventsMiddleware
from ..utils import generate_uuid
from prometheus_client import (
    Counter,
    Summary,
)
from typing import Optional


class DataPlane:
    """
    Internal implementation of handlers, used by both the gRPC and REST
    servers.
    """

    def __init__(self, settings: Settings, model_registry: MultiModelRegistry):
        self._settings = settings
        self._model_registry = model_registry

        self._inference_middleware = InferenceMiddlewares(
            CloudEventsMiddleware(settings)
        )

        self._ModelInferRequestSuccess = Counter(
            "model_infer_request_success",
            "Model infer request success count",
            ["model", "version"],
        )
        self._ModelInferRequestFailure = Counter(
            "model_infer_request_failure",
            "Model infer request failure count",
            ["model", "version"],
        )
        self._ModelInferRequestDuration = Summary(
            "model_infer_request_duration",
            "Model infer request duration",
            ["model", "version"],
        )

    async def live(self) -> bool:
        return True

    async def ready(self) -> bool:
        models = await self._model_registry.get_models()
        return all([model.ready for model in models])

    async def model_ready(self, name: str, version: Optional[str] = None) -> bool:
        model = await self._model_registry.get_model(name, version)
        return model.ready

    async def metadata(self) -> MetadataServerResponse:
        return MetadataServerResponse(
            name=self._settings.server_name,
            version=self._settings.server_version,
            extensions=self._settings.extensions,
        )

    async def model_metadata(
        self, name: str, version: Optional[str] = None
    ) -> MetadataModelResponse:
        model = await self._model_registry.get_model(name, version)
        # TODO: Make await optional for sync methods
        return await model.metadata()

    async def infer(
        self,
        payload: InferenceRequest,
        name: str,
        version: Optional[str] = None,
    ) -> InferenceResponse:
        infer_duration = self._ModelInferRequestDuration.labels(
            model=name, version=version
        ).time()
        infer_errors = self._ModelInferRequestFailure.labels(
            model=name, version=version
        ).count_exceptions()

        with infer_duration, infer_errors:
            if payload.id is None:
                payload.id = generate_uuid()

            model = await self._model_registry.get_model(name, version)

            self._inference_middleware.request_middleware(payload, model.settings)

            # TODO: Make await optional for sync methods
            prediction = await model.predict(payload)

            # Ensure ID matches
            prediction.id = payload.id

            self._inference_middleware.response_middleware(prediction, model.settings)

            self._ModelInferRequestSuccess.labels(model=name, version=version).inc()

            return prediction



from grpc import aio
from concurrent.futures import ThreadPoolExecutor
from typing import Optional, Any, List, Tuple

from ..handlers import DataPlane, ModelRepositoryHandlers
from ..settings import Settings

from .servicers import InferenceServicer
from .model_repository import ModelRepositoryServicer
from .dataplane_pb2_grpc import add_GRPCInferenceServiceServicer_to_server
from .model_repository_pb2_grpc import add_ModelRepositoryServiceServicer_to_server
from .interceptors import LoggingInterceptor, PromServerInterceptor
from .logging import logger

# Workers used for non-AsyncIO workloads (which aren't any in our case)
DefaultGrpcWorkers = 5


class GRPCServer:
    def __init__(
        self,
        settings: Settings,
        data_plane: DataPlane,
        model_repository_handlers: ModelRepositoryHandlers,
    ):
        self._settings = settings
        self._data_plane = data_plane
        self._model_repository_handlers = model_repository_handlers

    def _create_server(self):
        self._inference_servicer = InferenceServicer(
            self._data_plane, self._model_repository_handlers
        )
        self._model_repository_servicer = ModelRepositoryServicer(
            self._model_repository_handlers
        )

        interceptors = []

        if self._settings.debug:
            # If debug, enable access logs
            interceptors = [LoggingInterceptor()]

        if self._settings.metrics_endpoint:
            interceptors.append(PromServerInterceptor())

        self._server = aio.server(
            ThreadPoolExecutor(max_workers=DefaultGrpcWorkers),
            interceptors=tuple(interceptors),
            options=self._get_options(),
        )

        add_GRPCInferenceServiceServicer_to_server(
            self._inference_servicer, self._server
        )
        add_ModelRepositoryServiceServicer_to_server(
            self._model_repository_servicer, self._server
        )

        self._server.add_insecure_port(
            f"{self._settings.host}:{self._settings.grpc_port}"
        )

        return self._server

    def _get_options(self) -> List[Tuple[str, Any]]:
        options_dict = {}

        if self._settings._custom_grpc_server_settings:
            logger.warning(
                "gRPC custom configuration is out of support. Use as your own risk"
            )
            options_dict.update(self._settings._custom_grpc_server_settings)

        max_message_length = self._settings.grpc_max_message_length
        if max_message_length is not None:
            options_dict.update(
                {
                    "grpc.max_message_length": max_message_length,
                    "grpc.max_send_message_length": max_message_length,
                    "grpc.max_receive_message_length": max_message_length,
                }
            )

        return list(options_dict.items())

    async def start(self):
        self._create_server()

        await self._server.start()

        logger.info(
            "gRPC server running on "
            f"http://{self._settings.host}:{self._settings.grpc_port}"
        )
        await self._server.wait_for_termination()

    async def stop(self, sig: Optional[int] = None):
        logger.info("Waiting for gRPC server shutdown")
        # TODO: Read from config
        await self._server.stop(grace=5)
        logger.info("gRPC server shutdown complete")



from typing import Callable

from . import model_repository_pb2 as mr_pb
from .converters import (
    RepositoryIndexRequestConverter,
    RepositoryIndexResponseConverter,
)
from .model_repository_pb2_grpc import ModelRepositoryServiceServicer
from .utils import handle_mlserver_error
from .logging import logger
from ..handlers import ModelRepositoryHandlers


def _deprecated(f: Callable):
    async def _inner(self, request, context):
        logger.warning(
            "DEPRECATED!! "
            f"The `inference.model_repository.ModelRepositoryService/{f.__name__}` "
            "endpoint has now been deprecated and will be removed in "
            "MLServer 1.2.0. "
            f"Please use the `inference.GRPCInferenceService/{f.__name__}` "
            "endpoint instead."
        )
        return await f(self, request, context)

    return _inner


class ModelRepositoryServicer(ModelRepositoryServiceServicer):
    def __init__(self, handlers: ModelRepositoryHandlers):
        self._handlers = handlers

    @_deprecated
    async def RepositoryIndex(
        self, request: mr_pb.RepositoryIndexRequest, context
    ) -> mr_pb.RepositoryIndexResponse:
        payload = RepositoryIndexRequestConverter.to_types(request)
        index = await self._handlers.index(payload)
        return RepositoryIndexResponseConverter.from_types(  # type: ignore
            index, use_model_repository=True
        )

    @handle_mlserver_error
    @_deprecated
    async def RepositoryModelLoad(
        self, request: mr_pb.RepositoryModelLoadRequest, context
    ) -> mr_pb.RepositoryModelLoadResponse:
        await self._handlers.load(request.model_name)
        return mr_pb.RepositoryModelLoadResponse()

    @handle_mlserver_error
    @_deprecated
    async def RepositoryModelUnload(
        self, request: mr_pb.RepositoryModelUnloadRequest, context
    ) -> mr_pb.RepositoryModelUnloadResponse:
        await self._handlers.unload(request.model_name)
        return mr_pb.RepositoryModelUnloadResponse()



from .repository import ModelRepository, SchemalessModelRepository
from ..settings import Settings
from pydantic import PyObject


class ModelRepositoryFactory:
    @staticmethod
    def resolve_model_repository(settings: Settings) -> ModelRepository:
        model_repository_implementation: PyObject = SchemalessModelRepository

        result: ModelRepository
        if settings.model_repository_implementation:
            model_repository_implementation = settings.model_repository_implementation

        result = model_repository_implementation(
            root=settings.model_repository_root,
            **settings.model_repository_implementation_args,
        )

        return result



import asyncio
import signal
import logging

from typing import Optional, List

from mlserver.repository.factory import ModelRepositoryFactory

from .model import MLModel
from .settings import Settings, ModelSettings
from .logging import configure_logger
from .registry import MultiModelRegistry
from .handlers import DataPlane, ModelRepositoryHandlers
from .parallel import InferencePool
from .batching import load_batching
from .rest import RESTServer
from .grpc import GRPCServer
from .metrics import MetricsServer
from .kafka import KafkaServer
from .utils import logger

HANDLED_SIGNALS = [signal.SIGINT, signal.SIGTERM, signal.SIGQUIT]


class MLServer:
    def __init__(self, settings: Settings):
        self._settings = settings
        self._add_signal_handlers()

        self._metrics_server = None
        if self._settings.metrics_endpoint:
            self._metrics_server = MetricsServer(self._settings)

        self._inference_pool = None
        if self._settings.parallel_workers:
            # Only load inference pool if parallel inference has been enabled
            on_worker_stop = []
            if self._metrics_server:
                on_worker_stop = [self._metrics_server.on_worker_stop]

            self._inference_pool = InferencePool(
                self._settings, on_worker_stop=on_worker_stop  # type: ignore
            )

        self._model_registry = self._create_model_registry()
        self._model_repository = ModelRepositoryFactory.resolve_model_repository(
            self._settings
        )
        self._data_plane = DataPlane(
            settings=self._settings, model_registry=self._model_registry
        )
        self._model_repository_handlers = ModelRepositoryHandlers(
            repository=self._model_repository, model_registry=self._model_registry
        )

        self._configure_logger()
        self._create_servers()

    def _create_model_registry(self) -> MultiModelRegistry:
        on_model_load = [
            self.add_custom_handlers,
            load_batching,
        ]
        on_model_reload = [self.reload_custom_handlers]
        on_model_unload = [self.remove_custom_handlers]

        if self._inference_pool:
            on_model_load = [
                self._inference_pool.load_model,
                self.add_custom_handlers,
                load_batching,
            ]
            on_model_reload = [
                self._inference_pool.reload_model,  # type: ignore
                self.reload_custom_handlers,
            ]
            on_model_unload = [
                self._inference_pool.unload_model,  # type: ignore
                self.remove_custom_handlers,
            ]

        return MultiModelRegistry(
            on_model_load=on_model_load,  # type: ignore
            on_model_reload=on_model_reload,  # type: ignore
            on_model_unload=on_model_unload,  # type: ignore
        )

    def _configure_logger(self):
        logger.setLevel(logging.INFO)
        if self._settings.debug:
            logger.setLevel(logging.DEBUG)

        self._logger = configure_logger(self._settings)

    def _create_servers(self):
        self._rest_server = RESTServer(
            self._settings, self._data_plane, self._model_repository_handlers
        )
        self._grpc_server = GRPCServer(
            self._settings, self._data_plane, self._model_repository_handlers
        )

        self._kafka_server = None
        if self._settings.kafka_enabled:
            self._kafka_server = KafkaServer(self._settings, self._data_plane)

    async def start(self, models_settings: List[ModelSettings] = []):
        servers = [self._rest_server.start(), self._grpc_server.start()]
        if self._metrics_server:
            servers.append(self._metrics_server.start())

        if self._kafka_server:
            servers.append(self._kafka_server.start())

        servers_task = asyncio.gather(*servers)

        try:
            await asyncio.gather(
                *[
                    self._model_registry.load(model_settings)
                    for model_settings in models_settings
                ]
            )
        except Exception:
            # If one of the models failed to load during startup, shutdown the
            # server gracefully
            logger.exception("Some of the models failed to load during startup!")
            await self.stop()
        finally:
            await servers_task

    async def add_custom_handlers(self, model: MLModel) -> MLModel:
        await self._rest_server.add_custom_handlers(model)
        if self._kafka_server:
            await self._kafka_server.add_custom_handlers(model)

        # TODO: Add support for custom gRPC endpoints
        # self._grpc_server.add_custom_handlers(handlers)

        return model

    async def reload_custom_handlers(self, old_model: MLModel, new_model: MLModel):
        await self.add_custom_handlers(new_model)
        await self.remove_custom_handlers(old_model)

        # TODO: Add support for custom gRPC endpoints
        # self._grpc_server.delete_custom_handlers(handlers)

        return new_model

    async def remove_custom_handlers(self, model: MLModel):
        await self._rest_server.delete_custom_handlers(model)
        if self._kafka_server:
            await self._kafka_server.delete_custom_handlers(model)

        # TODO: Add support for custom gRPC endpoints
        # self._grpc_server.delete_custom_handlers(handlers)

        return model

    def _add_signal_handlers(self):
        loop = asyncio.get_event_loop()

        for sig in HANDLED_SIGNALS:
            loop.add_signal_handler(
                sig, lambda s=sig: asyncio.create_task(self.stop(sig=s))
            )

    async def stop(self, sig: Optional[int] = None):
        if self._inference_pool:
            await self._inference_pool.close()

        if self._kafka_server:
            await self._kafka_server.stop()

        if self._grpc_server:
            await self._grpc_server.stop(sig)

        if self._rest_server:
            await self._rest_server.stop(sig)

        if self._metrics_server:
            await self._metrics_server.stop(sig)



from functools import wraps
from typing import Awaitable, Callable, Optional

from ..errors import MLServerError
from ..model import MLModel
from ..types import InferenceRequest, InferenceResponse
from ..utils import get_wrapped_method
from .adaptive import AdaptiveBatcher

_AdaptiveBatchingAttr = "__adaptive_batching__"


class InvalidBatchingMethod(MLServerError):
    def __init__(self, method_name: str, reason: Optional[str] = None):
        msg = f"Method {method_name} can't be used for adaptive batching"
        if reason:
            msg += f": {reason}"

        super().__init__(msg)


def adaptive_batching(f: Callable[[InferenceRequest], Awaitable[InferenceResponse]]):
    """
    Decorator for the `predict()` method which will ensure it uses the
    underlying adaptive batcher instance.
    """

    @wraps(f)
    async def _inner(payload: InferenceRequest) -> InferenceResponse:
        wrapped_f = get_wrapped_method(f)
        if not hasattr(wrapped_f, "__self__"):
            raise InvalidBatchingMethod(
                wrapped_f.__name__, reason="method is not bound"
            )

        model = getattr(wrapped_f, "__self__")
        if not hasattr(model, _AdaptiveBatchingAttr):
            raise InvalidBatchingMethod(
                wrapped_f.__name__, reason="adaptive batching has not been loaded"
            )

        batcher = getattr(model, _AdaptiveBatchingAttr)
        return await batcher.predict(payload)

    return _inner


async def load_batching(model: MLModel) -> MLModel:
    if model.settings.max_batch_size <= 1:
        return model

    if model.settings.max_batch_time <= 0:
        return model

    batcher = AdaptiveBatcher(model)
    setattr(model, _AdaptiveBatchingAttr, batcher)

    # Decorate predict method
    setattr(model, "predict", adaptive_batching(model.predict))

    return model



"""
Command-line interface to manage MLServer models.
"""
import click
import asyncio

from functools import wraps

from .init_project import init_cookiecutter_project

from ..server import MLServer
from ..logging import logger, configure_logger
from ..utils import install_uvloop_event_loop

from .build import generate_dockerfile, build_image, write_dockerfile
from .serve import load_settings
from ..batch_processing import process_batch, CHOICES_TRANSPORT


def click_async(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        return asyncio.run(f(*args, **kwargs))

    return wrapper


@click.group()
@click.version_option()
def root():
    """
    Command-line interface to manage MLServer models.
    """
    pass


@root.command("start")
@click.argument("folder", nargs=1)
@click_async
async def start(folder: str):
    """
    Start serving a machine learning model with MLServer.
    """
    settings, models_settings = await load_settings(folder)

    server = MLServer(settings)
    await server.start(models_settings)


@root.command("build")
@click.argument("folder", nargs=1)
@click.option("-t", "--tag", type=str)
@click.option("--no-cache", default=False, is_flag=True)
@click_async
async def build(folder: str, tag: str, no_cache: bool = False):
    """
    Build a Docker image for a custom MLServer runtime.
    """
    dockerfile = generate_dockerfile()
    build_image(folder, dockerfile, tag, no_cache=no_cache)
    logger.info(f"Successfully built custom Docker image with tag {tag}")


@root.command("init")
# TODO: Update to have template(s) in the SeldonIO org
@click.option("-t", "--template", default="https://github.com/EthicalML/sml-security/")
@click_async
async def init_project(template: str):
    """
    Generate a base project template
    """
    init_cookiecutter_project(template)


@root.command("dockerfile")
@click.argument("folder", nargs=1)
@click.option("-i", "--include-dockerignore", is_flag=True)
@click_async
async def dockerfile(folder: str, include_dockerignore: bool):
    """
    Generate a Dockerfile
    """
    dockerfile = generate_dockerfile()
    dockerfile_path = write_dockerfile(
        folder, dockerfile, include_dockerignore=include_dockerignore
    )
    logger.info(f"Successfully written Dockerfile in {dockerfile_path}")


@root.command("infer")
@click.option(
    "--url",
    "-u",
    default="localhost:8080",
    envvar="MLSERVER_INFER_URL",
    help=(
        "URL of the MLServer to send inference requests to. "
        "Should not contain http or https."
    ),
)
@click.option(
    "--model-name",
    "-m",
    type=str,
    required=True,
    envvar="MLSERVER_INFER_MODEL_NAME",
    help="Name of the model to send inference requests to.",
)
@click.option(
    "--input-data-path",
    "-i",
    required=True,
    type=click.Path(),
    envvar="MLSERVER_INFER_INPUT_DATA_PATH",
    help="Local path to the input file containing inference requests to be processed.",
)
@click.option(
    "--output-data-path",
    "-o",
    required=True,
    type=click.Path(),
    envvar="MLSERVER_INFER_OUTPUT_DATA_PATH",
    help="Local path to the output file for the inference responses to be  written to.",
)
@click.option("--workers", "-w", default=10, envvar="MLSERVER_INFER_WORKERS")
@click.option("--retries", "-r", default=3, envvar="MLSERVER_INFER_RETRIES")
@click.option(
    "--batch-size",
    "-s",
    default=1,
    envvar="MLSERVER_INFER_BATCH_SIZE",
    help="Send inference requests grouped together as micro-batches.",
)
@click.option(
    "--binary-data",
    "-b",
    is_flag=True,
    default=False,
    envvar="MLSERVER_INFER_BINARY_DATA",
    help="Send inference requests as binary data (not fully supported).",
)
@click.option(
    "--verbose",
    "-v",
    is_flag=True,
    default=False,
    envvar="MLSERVER_INFER_VERBOSE",
    help="Verbose mode.",
)
@click.option(
    "--extra-verbose",
    "-vv",
    is_flag=True,
    default=False,
    envvar="MLSERVER_INFER_EXTRA_VERBOSE",
    help="Extra verbose mode (shows detailed requests and responses).",
)
@click.option(
    "--transport",
    "-t",
    envvar="MLSERVER_INFER_TRANSPORT",
    type=click.Choice(CHOICES_TRANSPORT),
    default="rest",
    help=(
        "Transport type to use to send inference requests. "
        "Can be 'rest' or 'grpc' (not yet supported)."
    ),
)
@click.option(
    "--request-headers",
    "-H",
    envvar="MLSERVER_INFER_REQUEST_HEADERS",
    type=str,
    multiple=True,
    help=(
        "Headers to be set on each inference request send to the server. "
        "Multiple options are allowed as: -H 'Header1: Val1' -H 'Header2: Val2'. "
        "When setting up as environmental provide as 'Header1:Val1 Header2:Val2'."
    ),
)
@click.option(
    "--timeout",
    default=60,
    envvar="MLSERVER_INFER_CONNECTION_TIMEOUT",
    help="Connection timeout to be passed to tritonclient.",
)
@click.option(
    "--batch-interval",
    default=0,
    type=float,
    envvar="MLSERVER_INFER_BATCH_INTERVAL",
    help="Minimum time interval (in seconds) between requests made by each worker.",
)
@click.option(
    "--batch-jitter",
    default=0,
    type=float,
    envvar="MLSERVER_INFER_BATCH_JITTER",
    help="Maximum random jitter (in seconds) added to batch interval between requests.",
)
@click.option(
    "--use-ssl",
    is_flag=True,
    default=False,
    envvar="MLSERVER_INFER_USE_SSL",
    help="Use SSL in communications with inference server.",
)
@click.option(
    "--insecure",
    is_flag=True,
    default=False,
    envvar="MLSERVER_INFER_INSECURE",
    help="Disable SSL verification in communications. Use with caution.",
)
@click_async
# TODO: add flags for SSL --key-file and --cert-file (see Tritonclient examples)
async def infer(
    model_name,
    url,
    workers,
    retries,
    batch_size,
    input_data_path,
    output_data_path,
    binary_data,
    transport,
    request_headers,
    timeout,
    batch_interval,
    batch_jitter,
    use_ssl,
    insecure,
    verbose,
    extra_verbose,
):
    """
    Execute batch inference requests against V2 inference server (experimental).
    """
    await process_batch(
        model_name=model_name,
        url=url,
        workers=workers,
        retries=retries,
        batch_size=batch_size,
        input_data_path=input_data_path,
        output_data_path=output_data_path,
        binary_data=binary_data,
        transport=transport,
        request_headers=request_headers,
        timeout=timeout,
        batch_interval=batch_interval,
        batch_jitter=batch_jitter,
        use_ssl=use_ssl,
        insecure=insecure,
        verbose=verbose,
        extra_verbose=extra_verbose,
    )


def main():
    configure_logger()
    install_uvloop_event_loop()
    root()


if __name__ == "__main__":
    main()



from typing import Callable

from fastapi import status
from fastapi.routing import APIRoute
from starlette.types import Scope

from ..handlers.custom import CustomHandler


def to_status_code(flag: bool, error_code: int = status.HTTP_400_BAD_REQUEST) -> int:
    """
    Convert a boolean flag into a HTTP status code.
    """
    if flag:
        return status.HTTP_200_OK

    return error_code


def matches(
    route: APIRoute, custom_handler: CustomHandler, handler_method: Callable
) -> bool:
    if route.endpoint != handler_method:
        return False

    scope = to_scope(custom_handler)
    match, _ = route.matches(scope)
    if match == match.NONE:
        return False

    return True


def to_scope(custom_handler: CustomHandler) -> Scope:
    return {
        "type": "http",
        "method": custom_handler.rest_method,
        "path": custom_handler.rest_path,
    }



import asyncio

from typing import Callable, Awaitable, List, Dict, Optional
from itertools import chain
from functools import cmp_to_key

from .model import MLModel
from .errors import ModelNotFound
from .logging import logger
from .settings import ModelSettings

ModelRegistryHook = Callable[[MLModel], Awaitable[MLModel]]
ModelReloadHook = Callable[[MLModel, MLModel], Awaitable[MLModel]]


def _get_version(model_settings: ModelSettings) -> Optional[str]:
    if model_settings.parameters:
        return model_settings.parameters.version

    return None


def _is_newer(a: MLModel, b: MLModel) -> int:
    """
    Returns true if 'a' is newer than 'b'.

    TODO: Support other ordering schemes (e.g. semver).
    """
    if a.version is None:
        return 1

    if b.version is None:
        return -1

    try:
        a_int = int(a.version)
        b_int = int(b.version)

        return a_int - b_int
    except ValueError:
        if a.version > b.version:
            return 1
        elif a.version < b.version:
            return -1
        else:
            return 0


class SingleModelRegistry:
    """
    Registry for a single model with multiple versions.
    """

    def __init__(
        self,
        model_settings: ModelSettings,
        on_model_load: List[ModelRegistryHook] = [],
        on_model_reload: List[ModelReloadHook] = [],
        on_model_unload: List[ModelRegistryHook] = [],
    ):
        self._versions: Dict[str, MLModel] = {}
        self._default: Optional[MLModel] = None

        self._name = model_settings.name
        self._on_model_load = on_model_load
        self._on_model_reload = on_model_reload
        self._on_model_unload = on_model_unload

    @property
    def default(self) -> Optional[MLModel]:
        if self._default is None:
            self._default = self._find_default()

        return self._default

    def _find_default(self) -> Optional[MLModel]:
        if self._default is None:
            if self._versions:
                version_key = cmp_to_key(_is_newer)
                latest_model = max(self._versions.values(), key=version_key)
                return latest_model

        return self._default

    def _clear_default(self):
        self._default = None

    def _refresh_default(
        self, new_model: Optional[MLModel] = None
    ) -> Optional[MLModel]:
        if new_model:
            # Check whether new model is "defaulter" than current default
            # NOTE: This should help to avoid iterating through all versioned
            # models each time a new model is loaded to find the latest

            if self._default is None:
                # If default is currently empty, take new one as new default
                self._default = new_model
                return new_model

            if new_model.version is None:
                # If new model doesn't have a version, assume it's "defaulter"
                # than previous default
                self._default = new_model
                return new_model

            if self._default.version is None:
                # If default doesn't have a version (and new one does), assume
                # that current default is "defaulter" than new one
                return self._default

            # Otherwise, compare versions
            if _is_newer(new_model, self._default) >= 0:
                self._default = new_model
                return new_model

            return self._default

        if self._default and self._default.version is None:
            # If there isn't a new model to compare, and current default has no
            # version, then consider that current one is "defaulter" than other
            # versioned models
            return self._default

        # Otherwise, find latest from current set of versions
        self._default = self._find_default()
        return self._default

    async def load(self, model_settings: ModelSettings) -> MLModel:
        # If there's a previously loaded model, we'll need to unload it at the
        # end
        previous_version = _get_version(model_settings)
        previous_loaded_model = self._find_model(previous_version)

        model_class = model_settings.implementation
        new_model = model_class(model_settings)  # type: ignore

        if previous_loaded_model:
            await self._reload_model(previous_loaded_model, new_model)
        else:
            await self._load_model(new_model)

        return new_model

    async def _load_model(self, model: MLModel):
        try:
            # Register the model before loading it, to ensure that the model
            # appears as a not-ready (i.e. loading) model
            self._register(model)

            for callback in self._on_model_load:
                # NOTE: Callbacks need to be executed sequentially to ensure that
                # they go in the right order
                model = await callback(model)

            # Register model again to ensure we save version modified by hooks
            self._register(model)
            await model.load()

            logger.info(f"Loaded model '{model.name}' succesfully.")
        except Exception:
            logger.info(
                f"Couldn't load model '{model.name}'. "
                "Model will be removed from registry."
            )
            await self._unload_model(model)
            raise

    async def _reload_model(self, old_model: MLModel, new_model: MLModel):
        for callback in self._on_model_reload:
            new_model = await callback(old_model, new_model)

        # Loading the model before unloading the old one - this will ensure
        # that at least one is available (sort of mimicking a rolling
        # deployment)
        await new_model.load()
        self._register(new_model)

        if old_model == self.default:
            self._clear_default()

        logger.info(f"Reloaded model '{new_model.name}' succesfully.")

    async def unload(self):
        models = await self.get_models()
        await asyncio.gather(*[self._unload_model(model) for model in models])

        self._versions.clear()
        self._clear_default()

        logger.info(f"Unloaded all versions of model '{self._name}' succesfully.")

    async def unload_version(self, version: Optional[str] = None):
        model = await self.get_model(version)
        await self._unload_model(model)

        model_msg = f"model '{model.name}'"
        if version:
            model_msg = f"version {version} of {model_msg}"

        logger.info(f"Unloaded {model_msg} succesfully.")

    async def _unload_model(self, model: MLModel):
        # NOTE: Every callback needs to run to ensure one doesn't block the
        # others
        await asyncio.gather(
            *[callback(model) for callback in self._on_model_unload],
            return_exceptions=True,
        )

        if model.version:
            del self._versions[model.version]

        if model == self.default:
            self._clear_default()

    def _find_model(self, version: Optional[str] = None) -> Optional[MLModel]:
        if version:
            if version not in self._versions:
                return None

            return self._versions[version]

        return self.default

    async def get_model(self, version: Optional[str] = None) -> MLModel:
        model = self._find_model(version)

        if model is None:
            raise ModelNotFound(self._name, version)

        return model

    async def get_models(self) -> List[MLModel]:
        # NOTE: `.values()` returns a "view" instead of a list
        models = list(self._versions.values())

        # Add default if not versioned (as it won't be present on the
        # `_versions` dict
        if self.default and not self.default.version:
            models.append(self.default)

        return models

    def _register(self, model: MLModel):
        if model.version:
            self._versions[model.version] = model

        self._refresh_default(model)

    def empty(self) -> bool:
        if self._versions:
            return False

        return self.default is None


class MultiModelRegistry:
    """
    Multiple model registry, where each model can have multiple versions.
    """

    def __init__(
        self,
        on_model_load: List[ModelRegistryHook] = [],
        on_model_reload: List[ModelReloadHook] = [],
        on_model_unload: List[ModelRegistryHook] = [],
    ):
        self._models: Dict[str, SingleModelRegistry] = {}
        self._on_model_load = on_model_load
        self._on_model_reload = on_model_reload
        self._on_model_unload = on_model_unload

    async def load(self, model_settings: ModelSettings) -> MLModel:
        if model_settings.name not in self._models:
            self._models[model_settings.name] = SingleModelRegistry(
                model_settings,
                on_model_load=self._on_model_load,
                on_model_reload=self._on_model_reload,
                on_model_unload=self._on_model_unload,
            )

        return await self._models[model_settings.name].load(model_settings)

    async def unload(self, name: str):
        model_registry = self._get_model_registry(name)
        await model_registry.unload()
        del self._models[name]

    async def unload_version(self, name: str, version: Optional[str] = None):
        model_registry = self._get_model_registry(name, version)
        await model_registry.unload_version(version)
        if model_registry.empty():
            del self._models[name]

    async def get_model(self, name: str, version: Optional[str] = None) -> MLModel:
        model_registry = self._get_model_registry(name, version)
        return await model_registry.get_model(version)

    async def get_models(self, name: Optional[str] = None) -> List[MLModel]:
        if name is not None:
            model_registry = self._get_model_registry(name)
            return await model_registry.get_models()

        models_list = await asyncio.gather(
            *[model.get_models() for model in self._models.values()]
        )

        return chain.from_iterable(models_list)  # type: ignore

    def _get_model_registry(
        self, name: str, version: Optional[str] = None
    ) -> SingleModelRegistry:
        if name not in self._models:
            raise ModelNotFound(name, version)

        return self._models[name]



from enum import Enum
from typing import Dict, Optional, Union

from .settings import Settings, ModelSettings
from .types import InferenceRequest, InferenceResponse, Parameters
from .middleware import InferenceMiddleware

CLOUDEVENTS_HEADER_ID = "Ce-Id"
CLOUDEVENTS_HEADER_SPECVERSION = "Ce-Specversion"
CLOUDEVENTS_HEADER_SOURCE = "Ce-Source"
CLOUDEVENTS_HEADER_TYPE = "Ce-Type"
CLOUDEVENTS_HEADER_SPECVERSION_DEFAULT = "0.3"

CLOUDEVENTS_HEADER_REQUEST_ID = "Ce-Requestid"
CLOUDEVENTS_HEADER_MODEL_ID = "Ce-Modelid"
CLOUDEVENTS_HEADER_INFERENCE_SERVICE = "Ce-Inferenceservicename"
CLOUDEVENTS_HEADER_NAMESPACE = "Ce-Namespace"
CLOUDEVENTS_HEADER_ENDPOINT = "Ce-Endpoint"


class CloudEventsTypes(Enum):
    Request = "io.seldon.serving.inference.request"
    Response = "io.seldon.serving.inference.response"


def get_namespace() -> Optional[str]:
    try:
        # Namespace can be fetched from loaded file vars from k8s 1.15.3+
        with open("/var/run/secrets/kubernetes.io/serviceaccount/namespace", "r") as f:
            return f.read()
    except FileNotFoundError:
        return None

    return None


def _update_headers(
    payload: Union[InferenceRequest, InferenceResponse], headers: Dict[str, str]
):
    if payload.parameters is None:
        payload.parameters = Parameters(headers=headers)
        return

    if payload.parameters.headers is None:
        payload.parameters.headers = headers
        return

    payload.parameters.headers.update(headers)


class CloudEventsMiddleware(InferenceMiddleware):
    def __init__(self, settings: Settings):
        self._settings = settings
        self._namespace = get_namespace()

    def _get_headers(
        self,
        ce_type: CloudEventsTypes,
        model_settings: ModelSettings,
        ce_id: Optional[str],
    ) -> Dict[str, str]:
        source = f"io.seldon.serving.deployment.{self._settings.server_name}"
        if self._namespace:
            source = f"{source}.{self._namespace}"

        ce_headers = {
            CLOUDEVENTS_HEADER_SPECVERSION: CLOUDEVENTS_HEADER_SPECVERSION_DEFAULT,
            CLOUDEVENTS_HEADER_SOURCE: source,
            CLOUDEVENTS_HEADER_TYPE: ce_type.value,
            CLOUDEVENTS_HEADER_MODEL_ID: model_settings.name,
            CLOUDEVENTS_HEADER_INFERENCE_SERVICE: self._settings.server_name,
            CLOUDEVENTS_HEADER_ENDPOINT: model_settings.name,
        }

        if ce_id:
            ce_headers[CLOUDEVENTS_HEADER_ID] = ce_id
            ce_headers[CLOUDEVENTS_HEADER_REQUEST_ID] = ce_id

        if self._namespace:
            ce_headers[CLOUDEVENTS_HEADER_NAMESPACE] = self._namespace

        return ce_headers

    def request_middleware(
        self,
        request: InferenceRequest,
        model_settings: ModelSettings,
    ) -> InferenceRequest:
        ce_headers = self._get_headers(
            CloudEventsTypes.Request, model_settings, request.id
        )

        _update_headers(request, ce_headers)
        return request

    def response_middleware(
        self,
        response: InferenceResponse,
        model_settings: ModelSettings,
    ) -> InferenceResponse:
        ce_headers = self._get_headers(
            CloudEventsTypes.Response, model_settings, response.id
        )

        _update_headers(response, ce_headers)
        return response



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from . import dataplane_pb2 as dataplane__pb2


class GRPCInferenceServiceStub(object):
    """
    Inference Server GRPC endpoints.

    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.ServerLive = channel.unary_unary(
            "/inference.GRPCInferenceService/ServerLive",
            request_serializer=dataplane__pb2.ServerLiveRequest.SerializeToString,
            response_deserializer=dataplane__pb2.ServerLiveResponse.FromString,
        )
        self.ServerReady = channel.unary_unary(
            "/inference.GRPCInferenceService/ServerReady",
            request_serializer=dataplane__pb2.ServerReadyRequest.SerializeToString,
            response_deserializer=dataplane__pb2.ServerReadyResponse.FromString,
        )
        self.ModelReady = channel.unary_unary(
            "/inference.GRPCInferenceService/ModelReady",
            request_serializer=dataplane__pb2.ModelReadyRequest.SerializeToString,
            response_deserializer=dataplane__pb2.ModelReadyResponse.FromString,
        )
        self.ServerMetadata = channel.unary_unary(
            "/inference.GRPCInferenceService/ServerMetadata",
            request_serializer=dataplane__pb2.ServerMetadataRequest.SerializeToString,
            response_deserializer=dataplane__pb2.ServerMetadataResponse.FromString,
        )
        self.ModelMetadata = channel.unary_unary(
            "/inference.GRPCInferenceService/ModelMetadata",
            request_serializer=dataplane__pb2.ModelMetadataRequest.SerializeToString,
            response_deserializer=dataplane__pb2.ModelMetadataResponse.FromString,
        )
        self.ModelInfer = channel.unary_unary(
            "/inference.GRPCInferenceService/ModelInfer",
            request_serializer=dataplane__pb2.ModelInferRequest.SerializeToString,
            response_deserializer=dataplane__pb2.ModelInferResponse.FromString,
        )
        self.RepositoryIndex = channel.unary_unary(
            "/inference.GRPCInferenceService/RepositoryIndex",
            request_serializer=dataplane__pb2.RepositoryIndexRequest.SerializeToString,
            response_deserializer=dataplane__pb2.RepositoryIndexResponse.FromString,
        )
        self.RepositoryModelLoad = channel.unary_unary(
            "/inference.GRPCInferenceService/RepositoryModelLoad",
            request_serializer=dataplane__pb2.RepositoryModelLoadRequest.SerializeToString,
            response_deserializer=dataplane__pb2.RepositoryModelLoadResponse.FromString,
        )
        self.RepositoryModelUnload = channel.unary_unary(
            "/inference.GRPCInferenceService/RepositoryModelUnload",
            request_serializer=dataplane__pb2.RepositoryModelUnloadRequest.SerializeToString,
            response_deserializer=dataplane__pb2.RepositoryModelUnloadResponse.FromString,
        )


class GRPCInferenceServiceServicer(object):
    """
    Inference Server GRPC endpoints.

    """

    def ServerLive(self, request, context):
        """Check liveness of the inference server."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def ServerReady(self, request, context):
        """Check readiness of the inference server."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def ModelReady(self, request, context):
        """Check readiness of a model in the inference server."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def ServerMetadata(self, request, context):
        """Get server metadata."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def ModelMetadata(self, request, context):
        """Get model metadata."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def ModelInfer(self, request, context):
        """Perform inference using a specific model."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def RepositoryIndex(self, request, context):
        """Get the index of model repository contents."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def RepositoryModelLoad(self, request, context):
        """Load or reload a model from a repository."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def RepositoryModelUnload(self, request, context):
        """Unload a model."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")


def add_GRPCInferenceServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
        "ServerLive": grpc.unary_unary_rpc_method_handler(
            servicer.ServerLive,
            request_deserializer=dataplane__pb2.ServerLiveRequest.FromString,
            response_serializer=dataplane__pb2.ServerLiveResponse.SerializeToString,
        ),
        "ServerReady": grpc.unary_unary_rpc_method_handler(
            servicer.ServerReady,
            request_deserializer=dataplane__pb2.ServerReadyRequest.FromString,
            response_serializer=dataplane__pb2.ServerReadyResponse.SerializeToString,
        ),
        "ModelReady": grpc.unary_unary_rpc_method_handler(
            servicer.ModelReady,
            request_deserializer=dataplane__pb2.ModelReadyRequest.FromString,
            response_serializer=dataplane__pb2.ModelReadyResponse.SerializeToString,
        ),
        "ServerMetadata": grpc.unary_unary_rpc_method_handler(
            servicer.ServerMetadata,
            request_deserializer=dataplane__pb2.ServerMetadataRequest.FromString,
            response_serializer=dataplane__pb2.ServerMetadataResponse.SerializeToString,
        ),
        "ModelMetadata": grpc.unary_unary_rpc_method_handler(
            servicer.ModelMetadata,
            request_deserializer=dataplane__pb2.ModelMetadataRequest.FromString,
            response_serializer=dataplane__pb2.ModelMetadataResponse.SerializeToString,
        ),
        "ModelInfer": grpc.unary_unary_rpc_method_handler(
            servicer.ModelInfer,
            request_deserializer=dataplane__pb2.ModelInferRequest.FromString,
            response_serializer=dataplane__pb2.ModelInferResponse.SerializeToString,
        ),
        "RepositoryIndex": grpc.unary_unary_rpc_method_handler(
            servicer.RepositoryIndex,
            request_deserializer=dataplane__pb2.RepositoryIndexRequest.FromString,
            response_serializer=dataplane__pb2.RepositoryIndexResponse.SerializeToString,
        ),
        "RepositoryModelLoad": grpc.unary_unary_rpc_method_handler(
            servicer.RepositoryModelLoad,
            request_deserializer=dataplane__pb2.RepositoryModelLoadRequest.FromString,
            response_serializer=dataplane__pb2.RepositoryModelLoadResponse.SerializeToString,
        ),
        "RepositoryModelUnload": grpc.unary_unary_rpc_method_handler(
            servicer.RepositoryModelUnload,
            request_deserializer=dataplane__pb2.RepositoryModelUnloadRequest.FromString,
            response_serializer=dataplane__pb2.RepositoryModelUnloadResponse.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
        "inference.GRPCInferenceService", rpc_method_handlers
    )
    server.add_generic_rpc_handlers((generic_handler,))


# This class is part of an EXPERIMENTAL API.
class GRPCInferenceService(object):
    """
    Inference Server GRPC endpoints.

    """

    @staticmethod
    def ServerLive(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/inference.GRPCInferenceService/ServerLive",
            dataplane__pb2.ServerLiveRequest.SerializeToString,
            dataplane__pb2.ServerLiveResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def ServerReady(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/inference.GRPCInferenceService/ServerReady",
            dataplane__pb2.ServerReadyRequest.SerializeToString,
            dataplane__pb2.ServerReadyResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def ModelReady(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/inference.GRPCInferenceService/ModelReady",
            dataplane__pb2.ModelReadyRequest.SerializeToString,
            dataplane__pb2.ModelReadyResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def ServerMetadata(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/inference.GRPCInferenceService/ServerMetadata",
            dataplane__pb2.ServerMetadataRequest.SerializeToString,
            dataplane__pb2.ServerMetadataResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def ModelMetadata(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/inference.GRPCInferenceService/ModelMetadata",
            dataplane__pb2.ModelMetadataRequest.SerializeToString,
            dataplane__pb2.ModelMetadataResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def ModelInfer(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/inference.GRPCInferenceService/ModelInfer",
            dataplane__pb2.ModelInferRequest.SerializeToString,
            dataplane__pb2.ModelInferResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def RepositoryIndex(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/inference.GRPCInferenceService/RepositoryIndex",
            dataplane__pb2.RepositoryIndexRequest.SerializeToString,
            dataplane__pb2.RepositoryIndexResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def RepositoryModelLoad(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/inference.GRPCInferenceService/RepositoryModelLoad",
            dataplane__pb2.RepositoryModelLoadRequest.SerializeToString,
            dataplane__pb2.RepositoryModelLoadResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def RepositoryModelUnload(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/inference.GRPCInferenceService/RepositoryModelUnload",
            dataplane__pb2.RepositoryModelUnloadRequest.SerializeToString,
            dataplane__pb2.RepositoryModelUnloadResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )



import grpc

from typing import Callable, Dict, Tuple
from fastapi import status

from grpc import ServicerContext

from .logging import logger
from ..errors import MLServerError


STATUS_CODE_MAPPING = {
    status.HTTP_400_BAD_REQUEST: grpc.StatusCode.INVALID_ARGUMENT,
    status.HTTP_404_NOT_FOUND: grpc.StatusCode.NOT_FOUND,
    status.HTTP_422_UNPROCESSABLE_ENTITY: grpc.StatusCode.FAILED_PRECONDITION,
    status.HTTP_500_INTERNAL_SERVER_ERROR: grpc.StatusCode.INTERNAL,
}


def to_headers(context: ServicerContext) -> Dict[str, str]:
    metadata = context.invocation_metadata()
    if hasattr(context, "trailing_metadata"):
        # NOTE: Older versions of `grpcio` (e.g. `grpcio==1.34.0`) don't expose
        # access to the trailing metadata on the service side
        metadata += context.trailing_metadata()
    headers = {}
    for metadatum in metadata:
        headers[metadatum.key] = metadatum.value

    return headers


def to_metadata(headers: Dict[str, str]) -> Tuple[Tuple[str, str], ...]:
    return tuple((key.lower(), value) for key, value in headers.items())


def _grpc_status_code(err: MLServerError):
    return STATUS_CODE_MAPPING.get(err.status_code, grpc.StatusCode.UNKNOWN)


def handle_mlserver_error(f: Callable):
    async def _inner(self, request, context):
        try:
            return await f(self, request, context)
        except MLServerError as err:
            logger.exception(err)
            await context.abort(code=_grpc_status_code(err), details=str(err))
        except Exception as err:
            logger.exception(err)
            await context.abort(code=grpc.StatusCode.INTERNAL, details=str(err))

    return _inner



from fastapi import status

from ..errors import MLServerError


class InvalidMessageHeaders(MLServerError):
    def __init__(self, missing_header: str):
        msg = (
            f"Invalid Kafka message. Expected '{missing_header}' header not "
            "found in message."
        )
        super().__init__(msg, status.HTTP_422_UNPROCESSABLE_ENTITY)



from typing import Any, Dict, Optional, List

from .codecs import (
    encode_response_output,
    encode_inference_response,
    decode_request_input,
    decode_inference_request,
    has_decoded,
    get_decoded,
    InputCodecLike,
    RequestCodecLike,
)
from .codecs.errors import CodecNotFound
from .settings import ModelSettings
from .types import (
    InferenceRequest,
    InferenceResponse,
    RequestInput,
    RequestOutput,
    ResponseOutput,
    MetadataModelResponse,
    MetadataTensor,
)
from .types import (
    Parameters,
)


def _generate_metadata_index(
    metadata_tensors: Optional[List[MetadataTensor]],
) -> Dict[str, MetadataTensor]:
    metadata_index: Dict[str, MetadataTensor] = {}

    if not metadata_tensors:
        return metadata_index

    for metadata_tensor in metadata_tensors:
        metadata_index[metadata_tensor.name] = metadata_tensor

    return metadata_index


class MLModel:
    """
    Abstract inference runtime which exposes the main interface to interact
    with ML models.
    """

    def __init__(self, settings: ModelSettings):
        self._settings = settings
        self._inputs_index: Dict[str, MetadataTensor] = {}

        self._inputs_index = _generate_metadata_index(self._settings.inputs)
        self._outputs_index = _generate_metadata_index(self._settings.outputs)

        self.ready = False

    async def load(self) -> bool:
        """
        Method responsible for loading the model from a model artefact.
        This method will be called on each of the parallel workers (when
        :doc:`parallel inference </user-guide/parallel-inference>`) is
        enabled).
        Its return value will represent the model's readiness status.
        A return value of ``True`` will mean the model is ready.

        **This method should be overriden to implement your custom load
        logic.**
        """
        self.ready = True
        return self.ready

    async def predict(self, payload: InferenceRequest) -> InferenceResponse:
        """
        Method responsible for running inference on the model.


        **This method should be overriden to implement your custom inference
        logic.**
        """
        raise NotImplementedError("predict() method not implemented")

    @property
    def name(self) -> str:
        """
        Model name, from the model settings.
        """
        return self._settings.name

    @property
    def version(self) -> Optional[str]:
        """
        Model version, from the model settings.
        """
        return self._settings.version

    @property
    def settings(self) -> ModelSettings:
        """
        Model settings.
        """
        return self._settings

    @property
    def inputs(self) -> Optional[List[MetadataTensor]]:
        """
        Expected model inputs, from the model settings.

        Note that this property can also be modified at model's load time to
        inject any inputs metadata.
        """
        return self._settings.inputs

    @inputs.setter
    def inputs(self, value: List[MetadataTensor]):
        self._settings.inputs = value
        self._inputs_index = _generate_metadata_index(self._settings.inputs)

    @property
    def outputs(self) -> Optional[List[MetadataTensor]]:
        """
        Expected model outputs, from the model settings.

        Note that this property can also be modified at model's load time to
        inject any outputs metadata.
        """
        return self._settings.outputs

    @outputs.setter
    def outputs(self, value: List[MetadataTensor]):
        self._settings.outputs = value
        self._outputs_index = _generate_metadata_index(self._settings.outputs)

    def decode(
        self,
        request_input: RequestInput,
        default_codec: Optional[InputCodecLike] = None,
    ) -> Any:
        """
        Helper to decode a **request input** into its corresponding high-level
        Python object.
        This method will find the most appropiate :doc:`input codec
        </user-guide/content-type>` based on the model's metadata and the
        input's content type.
        Otherwise, it will fall back to the codec specified in the
        ``default_codec`` kwarg.
        """
        decode_request_input(request_input, self._inputs_index)

        if has_decoded(request_input):
            return get_decoded(request_input)

        if default_codec:
            return default_codec.decode_input(request_input)

        return request_input.data

    def decode_request(
        self,
        inference_request: InferenceRequest,
        default_codec: Optional[RequestCodecLike] = None,
    ) -> Any:
        """
        Helper to decode an **inference request** into its corresponding
        high-level Python object.
        This method will find the most appropiate :doc:`request codec
        </user-guide/content-type>` based on the model's metadata and the
        requests's content type.
        Otherwise, it will fall back to the codec specified in the
        ``default_codec`` kwarg.
        """
        decode_inference_request(inference_request, self._settings, self._inputs_index)

        if has_decoded(inference_request):
            return get_decoded(inference_request)

        if default_codec:
            return default_codec.decode_request(inference_request)

        return inference_request

    def encode_response(
        self,
        payload: Any,
        default_codec: Optional[RequestCodecLike] = None,
    ) -> InferenceResponse:
        """
        Helper to encode a high-level Python object into its corresponding
        **inference response**.
        This method will find the most appropiate :doc:`request codec
        </user-guide/content-type>` based on the payload's type.
        Otherwise, it will fall back to the codec specified in the
        ``default_codec`` kwarg.
        """
        inference_response = encode_inference_response(payload, self._settings)

        if inference_response:
            return inference_response

        if default_codec:
            return default_codec.encode_response(self.name, payload, self.version)

        payload_type = str(type(payload))
        raise CodecNotFound(payload_type=payload_type, is_input=False, is_request=True)

    def encode(
        self,
        payload: Any,
        request_output: RequestOutput,
        default_codec: Optional[InputCodecLike] = None,
    ) -> ResponseOutput:
        """
        Helper to encode a high-level Python object into its corresponding
        **response output**.
        This method will find the most appropiate :doc:`input codec
        </user-guide/content-type>` based on the model's metadata, request
        output's content type or payload's type.
        Otherwise, it will fall back to the codec specified in the
        ``default_codec`` kwarg.
        """
        response_output = encode_response_output(
            payload, request_output, self._outputs_index
        )

        if response_output:
            return response_output

        if default_codec:
            return default_codec.encode_output(request_output.name, payload)

        raise CodecNotFound(name=request_output.name, is_input=False, is_request=False)

    async def metadata(self) -> MetadataModelResponse:
        model_metadata = MetadataModelResponse(
            name=self.name,
            platform=self._settings.platform,
            versions=self._settings.versions,
            inputs=self._settings.inputs,
            outputs=self._settings.outputs,
        )

        if self._settings.parameters:
            model_metadata.parameters = Parameters(
                content_type=self._settings.parameters.content_type
            )

        return model_metadata



DockerfileName = "Dockerfile"
DockerfileTemplate = """
FROM continuumio/miniconda3:4.12.0 AS env-builder
SHELL ["/bin/bash", "-c"]

ARG MLSERVER_ENV_NAME="mlserver-custom-env" \\
    MLSERVER_ENV_TARBALL="./envs/base.tar.gz"

RUN conda config --add channels conda-forge && \\
    conda install conda-pack

# The `[]` character range will ensure that Docker doesn't complain if the
# files don't exist:
# https://stackoverflow.com/a/65138098/5015573
COPY \\
    ./environment.ym[l] \\
    ./environment.yam[l] \\
    ./conda.ym[l] \\
    ./conda.yam[l] \\
    .
RUN mkdir $(dirname $MLSERVER_ENV_TARBALL); \\
    for envFile in environment.yml environment.yaml conda.yml conda.yaml; do \\
        if [[ -f $envFile ]]; then \\
            conda env create \
                --name $MLSERVER_ENV_NAME \\
                --file $envFile; \\
            conda-pack --ignore-missing-files \
                -n $MLSERVER_ENV_NAME \\
                -o $MLSERVER_ENV_TARBALL; \\
        fi \\
    done; \\
    chmod -R 776 $(dirname $MLSERVER_ENV_TARBALL)

FROM seldonio/mlserver:{version}-slim
SHELL ["/bin/bash", "-c"]

# Copy all potential sources for custom environments
COPY \\
    --chown=1000 \\
    --from=env-builder \\
    /envs/base.tar.g[z] \\
    ./envs/base.tar.gz
COPY \\
    ./settings.jso[n] \\
    ./model-settings.jso[n] \\
    ./requirements.tx[t] \\
    .

USER root
# Install dependencies system-wide, to ensure that they are available for every
# user and give permissions to (future) environment folder.
RUN ./hack/build-env.sh . && \\
    mkdir -p ./envs/base && \\
    chown -R 1000:0 ./envs/base && \\
    chmod -R 776 ./envs/base
USER 1000

# Copy everything else
COPY . .

# Override MLServer's own `CMD` to activate the embedded environment
# (optionally activating the hot-loaded one as well).
CMD source ./hack/activate-env.sh ./envs/base.tar.gz && \\
    mlserver start $MLSERVER_MODELS_DIR
"""

DockerignoreName = ".dockerignore"
Dockerignore = """
# Binaries for programs and plugins
*.exe
*.exe~
*.dll
*.so
*.dylib
*.pyc
*.pyo
*.pyd
bin

# Mac file system
**/.DS_Store

# Python dev
__pycache__
.Python
env
pip-log.txt
pip-delete-this-directory.txt
.mypy_cache
eggs/
.eggs/
*.egg-info/
./pytest_cache
.tox
build/
dist/

# Notebook Checkpoints
.ipynb_checkpoints

.coverage
.coverage.*
.cache
nosetests.xml
coverage.xml
*,cover
*.log
.git
"""



from functools import wraps, partial
from typing import (
    Any,
    Callable,
    Coroutine,
    Dict,
    List,
    Optional,
    Union,
    Type,
    Tuple,
    get_origin,
    get_args,
    get_type_hints,
    TYPE_CHECKING,
)


from ..types import InferenceRequest, InferenceResponse, ResponseOutput

from .base import RequestCodec, InputCodec, find_input_codec, find_request_codec
from .errors import InputsNotFound, OutputNotFound, CodecNotFound
from .utils import Codec

if TYPE_CHECKING:
    from ..model import MLModel

PredictFunc = Callable[
    ["MLModel", InferenceRequest], Coroutine[Any, Any, InferenceResponse]
]


def _as_list(a: Optional[Union[Any, Tuple[Any]]]) -> List[Any]:
    if a is None:
        return []

    if isinstance(a, tuple):
        # Split into components
        return list(a)

    # Otherwise, assume it's a single element
    return [a]


def _is_codec_type(c: Codec, t: Type) -> bool:
    if issubclass(c, t):  # type: ignore
        return True

    if isinstance(c, t):
        return True

    return False


_is_input_codec = partial(_is_codec_type, t=InputCodec)
_is_request_codec = partial(_is_codec_type, t=RequestCodec)


def _is_optional(t: Type) -> bool:
    origin = get_origin(t)
    if origin == Optional:
        return True

    if origin == Union:
        # Cover case where Optional[a] is reported as Union[a, None]
        args = get_args(t)
        if len(args) == 2 and type(None) in args:
            return True

    return False


def _unwrap_optional(t: Type) -> Type:
    args = get_args(t)
    for arg in args:
        if not isinstance(arg, type(None)):
            return arg

    return t


class SignatureCodec(RequestCodec):
    """
    Internal codec that knows how to map type hints to codecs.
    """

    # TODO: Should this receive the whole class as argument?
    def __init__(self, predict: Callable):
        self._predict = predict
        self._input_codecs, self._output_codecs = self._get_codecs(predict)

    def _get_codecs(self, pred: Callable) -> Tuple[Dict[str, Codec], List[Codec]]:
        self._input_hints = self._get_type_hints(pred)
        self._output_hints = _as_list(self._input_hints.pop("return", None))

        input_codecs = {}
        for name, type_hint in self._input_hints.items():
            codec = self._find_codec(name=name, type_hint=type_hint, is_input=True)
            # TODO: Consider metadata as well! (needs to be done at runtime)
            input_codecs[name] = codec

        output_codecs = []
        for type_hint in self._output_hints:
            # Try either as an input or as a request codec
            codec = self._find_codec(name=None, type_hint=type_hint, is_input=False)
            output_codecs.append(codec)

        return input_codecs, output_codecs

    def _get_type_hints(self, pred: Callable) -> Dict[str, Type]:
        type_hints = get_type_hints(pred)
        # For us, `typing.Optional` is just syntactic sugar, so let's ensure we
        # unwrap it
        for name, hint in type_hints.items():
            if _is_optional(hint):
                type_hints[name] = _unwrap_optional(hint)

        return type_hints

    def _find_codec(
        self, name: Optional[str], type_hint: Type, is_input: bool = False
    ) -> Codec:
        codec = find_input_codec(type_hint=type_hint)
        if codec is not None:
            return codec

        codec = find_request_codec(type_hint=type_hint)  # type: ignore
        if codec is not None:
            return codec

        raise CodecNotFound(name=name, payload_type=str(type_hint), is_input=is_input)

    def decode_request(  # type: ignore
        self, request: InferenceRequest
    ) -> Dict[str, Any]:
        inputs = {}
        extra_request_inputs = []
        for request_input in request.inputs:
            input_name = request_input.name
            if input_name not in self._input_codecs:
                # Aggregate extra request inputs to check later, as they could
                # be part of aggregated request codecs (e.g. like dataframes)
                extra_request_inputs.append(request_input)
                continue

            # Ensure matching codec is an input codec
            codec = self._input_codecs[input_name]
            if not _is_input_codec(codec):
                raise CodecNotFound(name=input_name, is_input=True)

            inputs[input_name] = codec.decode_input(request_input)  # type: ignore

        if extra_request_inputs:
            request_codec = self._get_request_codec()
            if not request_codec:
                # If there are no request codecs that can aggregate all
                # remaining inputs, raise an error
                raise InputsNotFound(extra_request_inputs, self._input_codecs)

            # We create a fake request built from the extra request inputs
            name, codec = request_codec
            extra_inputs = InferenceRequest(inputs=extra_request_inputs)
            inputs[name] = codec.decode_request(extra_inputs)

        return inputs

    def _get_request_codec(self) -> Optional[Tuple[str, RequestCodec]]:
        for name, codec in self._input_codecs.items():
            if _is_request_codec(codec):
                return name, codec  # type: ignore

        return None

    def encode_response(  # type: ignore
        self, model_name: str, payload: Any, model_version: Optional[str] = None
    ) -> InferenceResponse:
        payloads = _as_list(payload)
        outputs = []
        for idx, payload in enumerate(payloads):
            outputs += self._encode_outputs(idx, payload)

        return InferenceResponse(
            model_name=model_name, model_version=model_version, outputs=outputs
        )

    def _encode_outputs(self, idx: int, payload: Any) -> List[ResponseOutput]:
        output_type = type(payload)
        if idx >= len(self._output_codecs):
            raise OutputNotFound(idx, output_type, self._output_hints)

        # TODO: Fallback to encode_by_payload?
        codec = self._output_codecs[idx]
        if not codec.can_encode(payload):
            raise OutputNotFound(idx, output_type, self._output_hints)

        if _is_input_codec(codec):
            # TODO: Check model metadata for output names
            output_name = f"output-{idx}"
            response_output = codec.encode_output(  # type: ignore
                name=output_name, payload=payload
            )
            return [response_output]

        if _is_request_codec(codec):
            # NOTE: We will ignore `model_name` and only grab the outputs
            response = codec.encode_response(  # type: ignore
                model_name="", payload=payload
            )
            return response.outputs

        return []


def decode_args(predict: Callable) -> PredictFunc:
    codec = SignatureCodec(predict)

    @wraps(predict)
    async def _f(self: "MLModel", request: InferenceRequest) -> InferenceResponse:
        inputs = codec.decode_request(request=request)

        outputs = await predict(self, **inputs)

        return codec.encode_response(
            model_name=self.name, payload=outputs, model_version=self.version
        )

    return _f



from typing import Optional, Any, Dict, List, Type

from ..types import RequestInput
from ..errors import MLServerError


class CodecNotFound(MLServerError):
    def __init__(
        self,
        name: Optional[str] = None,
        payload_type: Optional[str] = None,
        is_input: bool = False,
        is_request: bool = False,
    ):
        msg = ""
        if name:
            msg = f"with name '{name}'"

            if payload_type:
                msg = f"{msg} and type {payload_type}"
        elif payload_type:
            msg = f"with type {payload_type}"

        field_category = ""
        if is_input:
            if is_request:
                field_category = "input request"
            else:
                field_category = "input field"
        else:
            if is_request:
                field_category = "output response"
            else:
                field_category = "output field"

        msg = f"Codec not found for {field_category} {msg}"
        super().__init__(msg)


class CodecError(MLServerError):
    def __init__(self, msg: str):
        msg = f"There was an error encoding / decoding the payload: {msg}"
        super().__init__(msg)


class OutputNotFound(MLServerError):
    def __init__(self, output_idx: int, output_type: Type, output_hints: List[Type]):
        expected_outputs = [f"'{output_hint}'" for output_hint in output_hints]
        msg = (
            f"Unexpected output value at position '{output_idx}' ({output_type}). "
            f"Expected outputs are {', '.join(expected_outputs)} outputs."
        )
        super().__init__(msg)


class InputsNotFound(MLServerError):
    def __init__(self, inputs: List[RequestInput], input_hints: Dict[str, Any]):
        input_names = [f"'{inp.name}'" for inp in inputs]
        available_inputs = [f"'{input_name}'" for input_name in input_hints.keys()]
        msg = (
            f"Input {', '.join(input_names)} was not found. "
            f"Available inputs are {', '.join(available_inputs)}."
        )
        super().__init__(msg)



import uvicorn
import os

from typing import Optional, TYPE_CHECKING

from fastapi import FastAPI

from ..settings import Settings
from .logging import logger
from .prometheus import PrometheusEndpoint, stop_metrics

if TYPE_CHECKING:
    from ..parallel import Worker


class _NoSignalServer(uvicorn.Server):
    def install_signal_handlers(self):
        pass


class MetricsServer:
    def __init__(self, settings: Settings):
        self._settings = settings
        self._endpoint = PrometheusEndpoint(settings)
        self._app = self._get_app()

    def _get_app(self):
        app = FastAPI(debug=self._settings.debug)
        app.add_route(self._settings.metrics_endpoint, self._endpoint.handle_metrics)
        return app

    async def on_worker_stop(self, worker: "Worker") -> None:
        if not worker.pid:
            return

        # NOTE: If a worker gets restarted (instead of regular shutdown), we may
        # not want to remove old files to keep counts intact
        await stop_metrics(self._settings, worker.pid)

    async def start(self):
        cfg = self._get_config()
        self._server = _NoSignalServer(cfg)

        metrics_server = f"http://{self._settings.host}:{self._settings.metrics_port}"
        logger.info(f"Metrics server running on {metrics_server}")
        logger.info(
            "Prometheus scraping endpoint can be accessed on "
            f"{metrics_server}{self._settings.metrics_endpoint}"
        )
        await self._server.serve()

    def _get_config(self):
        kwargs = {}

        if self._settings._custom_metrics_server_settings:
            logger.warning(
                "REST custom configuration is out of support. Use as your own risk"
            )
            kwargs.update(self._settings._custom_metrics_server_settings)

        kwargs.update(
            {
                "host": self._settings.host,
                "port": self._settings.metrics_port,
                "access_log": self._settings.debug,
            }
        )

        if self._settings.logging_settings:
            # If not None, use ours. Otherwise, let Uvicorn fall back on its
            # own config.
            kwargs.update({"log_config": self._settings.logging_settings})

        return uvicorn.Config(self._app, **kwargs)

    async def stop(self, sig: Optional[int] = None):
        await stop_metrics(self._settings, os.getpid())
        self._server.handle_exit(sig=sig, frame=None)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: model_repository.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x16model_repository.proto\x12\x1ainference.model_repository"@\n\x16RepositoryIndexRequest\x12\x17\n\x0frepository_name\x18\x01 \x01(\t\x12\r\n\x05ready\x18\x02 \x01(\x08"\xb5\x01\n\x17RepositoryIndexResponse\x12N\n\x06models\x18\x01 \x03(\x0b\x32>.inference.model_repository.RepositoryIndexResponse.ModelIndex\x1aJ\n\nModelIndex\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07version\x18\x02 \x01(\t\x12\r\n\x05state\x18\x03 \x01(\t\x12\x0e\n\x06reason\x18\x04 \x01(\t"I\n\x1aRepositoryModelLoadRequest\x12\x17\n\x0frepository_name\x18\x01 \x01(\t\x12\x12\n\nmodel_name\x18\x02 \x01(\t"\x1d\n\x1bRepositoryModelLoadResponse"K\n\x1cRepositoryModelUnloadRequest\x12\x17\n\x0frepository_name\x18\x01 \x01(\t\x12\x12\n\nmodel_name\x18\x02 \x01(\t"\x1f\n\x1dRepositoryModelUnloadResponse2\xb2\x03\n\x16ModelRepositoryService\x12|\n\x0fRepositoryIndex\x12\x32.inference.model_repository.RepositoryIndexRequest\x1a\x33.inference.model_repository.RepositoryIndexResponse"\x00\x12\x88\x01\n\x13RepositoryModelLoad\x12\x36.inference.model_repository.RepositoryModelLoadRequest\x1a\x37.inference.model_repository.RepositoryModelLoadResponse"\x00\x12\x8e\x01\n\x15RepositoryModelUnload\x12\x38.inference.model_repository.RepositoryModelUnloadRequest\x1a\x39.inference.model_repository.RepositoryModelUnloadResponse"\x00\x62\x06proto3'
)


_REPOSITORYINDEXREQUEST = DESCRIPTOR.message_types_by_name["RepositoryIndexRequest"]
_REPOSITORYINDEXRESPONSE = DESCRIPTOR.message_types_by_name["RepositoryIndexResponse"]
_REPOSITORYINDEXRESPONSE_MODELINDEX = _REPOSITORYINDEXRESPONSE.nested_types_by_name[
    "ModelIndex"
]
_REPOSITORYMODELLOADREQUEST = DESCRIPTOR.message_types_by_name[
    "RepositoryModelLoadRequest"
]
_REPOSITORYMODELLOADRESPONSE = DESCRIPTOR.message_types_by_name[
    "RepositoryModelLoadResponse"
]
_REPOSITORYMODELUNLOADREQUEST = DESCRIPTOR.message_types_by_name[
    "RepositoryModelUnloadRequest"
]
_REPOSITORYMODELUNLOADRESPONSE = DESCRIPTOR.message_types_by_name[
    "RepositoryModelUnloadResponse"
]
RepositoryIndexRequest = _reflection.GeneratedProtocolMessageType(
    "RepositoryIndexRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _REPOSITORYINDEXREQUEST,
        "__module__": "model_repository_pb2"
        # @@protoc_insertion_point(class_scope:inference.model_repository.RepositoryIndexRequest)
    },
)
_sym_db.RegisterMessage(RepositoryIndexRequest)

RepositoryIndexResponse = _reflection.GeneratedProtocolMessageType(
    "RepositoryIndexResponse",
    (_message.Message,),
    {
        "ModelIndex": _reflection.GeneratedProtocolMessageType(
            "ModelIndex",
            (_message.Message,),
            {
                "DESCRIPTOR": _REPOSITORYINDEXRESPONSE_MODELINDEX,
                "__module__": "model_repository_pb2"
                # @@protoc_insertion_point(class_scope:inference.model_repository.RepositoryIndexResponse.ModelIndex)
            },
        ),
        "DESCRIPTOR": _REPOSITORYINDEXRESPONSE,
        "__module__": "model_repository_pb2"
        # @@protoc_insertion_point(class_scope:inference.model_repository.RepositoryIndexResponse)
    },
)
_sym_db.RegisterMessage(RepositoryIndexResponse)
_sym_db.RegisterMessage(RepositoryIndexResponse.ModelIndex)

RepositoryModelLoadRequest = _reflection.GeneratedProtocolMessageType(
    "RepositoryModelLoadRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _REPOSITORYMODELLOADREQUEST,
        "__module__": "model_repository_pb2"
        # @@protoc_insertion_point(class_scope:inference.model_repository.RepositoryModelLoadRequest)
    },
)
_sym_db.RegisterMessage(RepositoryModelLoadRequest)

RepositoryModelLoadResponse = _reflection.GeneratedProtocolMessageType(
    "RepositoryModelLoadResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _REPOSITORYMODELLOADRESPONSE,
        "__module__": "model_repository_pb2"
        # @@protoc_insertion_point(class_scope:inference.model_repository.RepositoryModelLoadResponse)
    },
)
_sym_db.RegisterMessage(RepositoryModelLoadResponse)

RepositoryModelUnloadRequest = _reflection.GeneratedProtocolMessageType(
    "RepositoryModelUnloadRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _REPOSITORYMODELUNLOADREQUEST,
        "__module__": "model_repository_pb2"
        # @@protoc_insertion_point(class_scope:inference.model_repository.RepositoryModelUnloadRequest)
    },
)
_sym_db.RegisterMessage(RepositoryModelUnloadRequest)

RepositoryModelUnloadResponse = _reflection.GeneratedProtocolMessageType(
    "RepositoryModelUnloadResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _REPOSITORYMODELUNLOADRESPONSE,
        "__module__": "model_repository_pb2"
        # @@protoc_insertion_point(class_scope:inference.model_repository.RepositoryModelUnloadResponse)
    },
)
_sym_db.RegisterMessage(RepositoryModelUnloadResponse)

_MODELREPOSITORYSERVICE = DESCRIPTOR.services_by_name["ModelRepositoryService"]
if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    _REPOSITORYINDEXREQUEST._serialized_start = 54
    _REPOSITORYINDEXREQUEST._serialized_end = 118
    _REPOSITORYINDEXRESPONSE._serialized_start = 121
    _REPOSITORYINDEXRESPONSE._serialized_end = 302
    _REPOSITORYINDEXRESPONSE_MODELINDEX._serialized_start = 228
    _REPOSITORYINDEXRESPONSE_MODELINDEX._serialized_end = 302
    _REPOSITORYMODELLOADREQUEST._serialized_start = 304
    _REPOSITORYMODELLOADREQUEST._serialized_end = 377
    _REPOSITORYMODELLOADRESPONSE._serialized_start = 379
    _REPOSITORYMODELLOADRESPONSE._serialized_end = 408
    _REPOSITORYMODELUNLOADREQUEST._serialized_start = 410
    _REPOSITORYMODELUNLOADREQUEST._serialized_end = 485
    _REPOSITORYMODELUNLOADRESPONSE._serialized_start = 487
    _REPOSITORYMODELUNLOADRESPONSE._serialized_end = 518
    _MODELREPOSITORYSERVICE._serialized_start = 521
    _MODELREPOSITORYSERVICE._serialized_end = 955
# @@protoc_insertion_point(module_scope)



from typing import Optional
from fastapi import Request
from pydantic import BaseModel

from .responses import Response
from ..errors import MLServerError


class APIErrorResponse(BaseModel):
    error: Optional[str] = None


async def handle_mlserver_error(request: Request, exc: MLServerError) -> Response:
    err_res = APIErrorResponse(error=str(exc))
    return Response(status_code=exc.status_code, content=err_res.dict())


_EXCEPTION_HANDLERS = {MLServerError: handle_mlserver_error}



from typing import Any, List

from ..types import RequestInput, ResponseOutput, Parameters

from .utils import SingleInputRequestCodec, InputOrOutput
from .base import InputCodec, register_input_codec, register_request_codec
from .lists import as_list, is_list_of, ListElement

_DefaultStrCodec = "utf-8"


def encode_str(elem: str) -> bytes:
    return elem.encode(_DefaultStrCodec)


def decode_str(encoded: ListElement, str_codec=_DefaultStrCodec) -> str:
    if encoded is None:
        return None

    if isinstance(encoded, bytes):
        return encoded.decode(str_codec)

    if isinstance(encoded, str):
        # NOTE: It may be a string already when decoded from json
        return encoded

    # TODO: Should we raise an error here?
    return ""


def _decode_input_or_output(input_or_output: InputOrOutput) -> List[str]:
    packed = input_or_output.data.__root__
    unpacked = map(decode_str, as_list(packed))
    return list(unpacked)


@register_input_codec
class StringCodec(InputCodec):
    """
    Encodes a list of Python strings as a BYTES input (output).
    """

    ContentType = "str"
    TypeHint = List[str]

    @classmethod
    def can_encode(cls, payload: Any) -> bool:
        return is_list_of(payload, str)

    @classmethod
    def encode_output(
        cls, name: str, payload: List[str], use_bytes: bool = True, **kwargs
    ) -> ResponseOutput:
        packed = payload
        if use_bytes:
            packed = list(map(encode_str, payload))  # type: ignore

        shape = [len(payload), 1]
        return ResponseOutput(
            name=name,
            datatype="BYTES",
            shape=shape,
            data=list(packed),
            parameters=Parameters(content_type=cls.ContentType),
        )

    @classmethod
    def decode_output(cls, response_output: ResponseOutput) -> List[str]:
        return _decode_input_or_output(response_output)

    @classmethod
    def decode_input(cls, request_input: RequestInput) -> List[str]:
        return _decode_input_or_output(request_input)

    @classmethod
    def encode_input(
        cls, name: str, payload: List[str], use_bytes: bool = True, **kwargs
    ) -> RequestInput:
        output = cls.encode_output(name, payload, use_bytes)

        return RequestInput(
            name=output.name,
            datatype=output.datatype,
            shape=output.shape,
            data=output.data,
            parameters=output.parameters,
        )


@register_request_codec
class StringRequestCodec(SingleInputRequestCodec):
    """
    Decodes the first input (output) of request (response) as a list of
    strings.
    This codec can be useful for cases where the whole payload is a single
    list of strings.
    """

    InputCodec = StringCodec
    ContentType = StringCodec.ContentType
    TypeHint = List[str]



import logging


class HealthEndpointFilter(logging.Filter):
    """
    Filter to avoid logging health endpoints.
    From:
        https://github.com/encode/starlette/issues/864#issuecomment-653076434
    """

    def filter(self, record: logging.LogRecord) -> bool:
        if not isinstance(record.args, tuple):
            return True

        if len(record.args) < 3:
            return True

        request_method = record.args[1]
        query_string = record.args[2]
        if request_method != "GET":
            return True

        if query_string in ["/v2/health/live", "/v2/health/ready"]:
            return False

        return True


def disable_health_access_logs() -> None:
    uvicorn_logger = logging.getLogger("uvicorn.access")
    uvicorn_logger.addFilter(HealthEndpointFilter())


loggerName = "mlserver.rest"
logger = logging.getLogger(loggerName)



from fastapi import status
from typing import Optional


class MLServerError(Exception):
    def __init__(self, msg: str, status_code: int = status.HTTP_400_BAD_REQUEST):
        super().__init__(msg)
        self.status_code = status_code


class InvalidModelURI(MLServerError):
    def __init__(self, name: str, model_uri: Optional[str] = None):
        msg = f"Invalid URI specified for model {name}"
        if model_uri:
            msg += f" ({model_uri})"

        super().__init__(msg, status.HTTP_422_UNPROCESSABLE_ENTITY)


class ModelNotFound(MLServerError):
    def __init__(self, name: str, version: Optional[str] = None):
        msg = f"Model {name} not found"
        if version is not None:
            msg = f"Model {name} with version {version} not found"

        super().__init__(msg, status.HTTP_404_NOT_FOUND)


class InferenceError(MLServerError):
    def __init__(self, msg: str):
        super().__init__(msg, status.HTTP_400_BAD_REQUEST)


class ModelParametersMissing(MLServerError):
    def __init__(self, model_name: str):
        super().__init__(
            f"Parameters missing for model {model_name}", status.HTTP_400_BAD_REQUEST
        )



import numpy as np

from typing import Any

from ..types import RequestInput, ResponseOutput, Parameters

from .base import InputCodec, register_input_codec, register_request_codec
from .utils import SingleInputRequestCodec, InputOrOutput, inject_batch_dimension
from .lists import is_list_of
from .string import encode_str

_DatatypeToNumpy = {
    "BOOL": "bool",
    "UINT8": "uint8",
    "UINT16": "uint16",
    "UINT32": "uint32",
    "UINT64": "uint64",
    "INT8": "int8",
    "INT16": "int16",
    "INT32": "int32",
    "INT64": "int64",
    "FP16": "float16",
    "FP32": "float32",
    "FP64": "float64",
    "BYTES": "bytes",
}

_NumpyToDatatype = {value: key for key, value in _DatatypeToNumpy.items()}

# NOTE: numpy has more types than v2 protocol
_NumpyToDatatype["object"] = "BYTES"
_NumpyToDatatype["S"] = "BYTES"
_NumpyToDatatype["U"] = "BYTES"


def to_dtype(input_or_output: InputOrOutput) -> "np.dtype":
    dtype = _DatatypeToNumpy[input_or_output.datatype]

    if input_or_output.datatype == "BYTES":
        data = getattr(input_or_output.data, "__root__", input_or_output.data)
        if is_list_of(data, str):
            # Handle special case of strings being treated as Numpy arrays
            return np.dtype(str)

        # bytes have variable size, so need to specify as part of type
        # TODO: Make elem size variable (and not just the last dimension)
        elem_size = input_or_output.shape[-1]
        return np.dtype((dtype, elem_size))

    return np.dtype(dtype)


def to_datatype(dtype: np.dtype) -> str:
    as_str = str(dtype)

    if as_str not in _NumpyToDatatype:
        # If not present, try with kind
        as_str = getattr(dtype, "kind")

    datatype = _NumpyToDatatype[as_str]

    return datatype


def _to_ndarray(input_or_output: InputOrOutput) -> np.ndarray:
    data = getattr(input_or_output.data, "__root__", input_or_output.data)
    dtype = to_dtype(input_or_output)

    if input_or_output.datatype == "BYTES":
        if is_list_of(data, bytes):
            # If the inputs is of type `BYTES`, there could be multiple "lists"
            # serialised into multiple buffers.
            # We will deserialise all of them and concatenate them together.
            decoded = [np.frombuffer(buffer, dtype) for buffer in data]
            return np.concatenate(decoded)

    return np.array(data, dtype)


def _encode_data(data: np.ndarray, datatype: str) -> list:
    if datatype == "BYTES":
        if np.issubdtype(data.dtype, str):
            # Handle special case of a string Numpy array, where the diff elems
            # need to be encoded as well
            as_list = data.flatten().tolist()
            return list(map(encode_str, as_list))

        if np.issubdtype(data.dtype, bytes):
            # `tobytes` is way faster than tolist, although it's harder to serialise
            # and only makes sense for actual bytes inputs (#253).
            # Note that `.tobytes()` will return a single `bytes` payload, thus we
            # need to encapsulate it into a list so that it's compatible.
            return [data.tobytes()]

    return data.flatten().tolist()


@register_input_codec
class NumpyCodec(InputCodec):
    """
    Decodes an request input (response output) as a NumPy array.
    """

    ContentType = "np"
    TypeHint = np.ndarray

    @classmethod
    def can_encode(csl, payload: Any) -> bool:
        return isinstance(payload, np.ndarray)

    @classmethod
    def encode_output(cls, name: str, payload: np.ndarray, **kwargs) -> ResponseOutput:
        datatype = to_datatype(payload.dtype)

        shape = inject_batch_dimension(list(payload.shape))

        return ResponseOutput(
            name=name,
            datatype=datatype,
            shape=shape,
            data=_encode_data(payload, datatype),
        )

    @classmethod
    def decode_output(cls, response_output: ResponseOutput) -> np.ndarray:
        return cls.decode_input(response_output)  # type: ignore

    @classmethod
    def encode_input(cls, name: str, payload: np.ndarray, **kwargs) -> RequestInput:
        output = cls.encode_output(name=name, payload=payload)

        return RequestInput(
            name=output.name,
            datatype=output.datatype,
            shape=output.shape,
            data=output.data,
            parameters=Parameters(content_type=cls.ContentType),
        )

    @classmethod
    def decode_input(cls, request_input: RequestInput) -> np.ndarray:
        model_data = _to_ndarray(request_input)

        # TODO: Check if reshape not valid
        return model_data.reshape(request_input.shape)


@register_request_codec
class NumpyRequestCodec(SingleInputRequestCodec):
    """
    Decodes the first input (output) of request (response) as a NumPy array.
    This codec can be useful for cases where the whole payload is a single
    NumPy tensor.
    """

    InputCodec = NumpyCodec
    ContentType = NumpyCodec.ContentType



import json

from typing import Any

from starlette.responses import JSONResponse as _JSONResponse

from ..codecs.string import decode_str

try:
    import orjson
except ImportError:
    orjson = None  # type: ignore


class BytesJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, bytes):
            # If we get a bytes payload, try to decode it back to a string on a
            # "best effort" basis
            return decode_str(obj)

        return super().default(self, obj)


class Response(_JSONResponse):
    """
    Custom Response class to use `orjson` if present.
    Otherwise, it'll fall back to the standard JSONResponse.
    """

    media_type = "application/json"

    def render(self, content: Any) -> bytes:
        if orjson is None:
            # Original implementation of starlette's JSONResponse, using our
            # custom encoder (capable of "encoding" bytes).
            # Original implementation can be seen here:
            # https://github.com/encode/starlette/blob/
            # f53faba229e3fa2844bc3753e233d9c1f54cca52/starlette/responses.py#L173-L180
            return json.dumps(
                content,
                ensure_ascii=False,
                allow_nan=False,
                indent=None,
                separators=(",", ":"),
                cls=BytesJSONEncoder,
            ).encode("utf-8")

        # This is equivalent to the ORJSONResponse implementation in FastAPI:
        # https://github.com/tiangolo/fastapi/blob/
        # 864643ef7608d28ac4ed321835a7fb4abe3dfc13/fastapi/responses.py#L32-L34
        return orjson.dumps(content, default=_encode_bytes)


def _encode_bytes(obj: Any) -> str:
    """
    Add compatibility with `bytes` payloads to `orjson`
    """
    if isinstance(obj, bytes):
        # If we get a bytes payload, try to decode it back to a string on a
        # "best effort" basis
        return decode_str(obj)

    raise TypeError



from typing import Any
from gzip import decompress as gzip_decompress
from fastapi import Request as _Request

try:
    import orjson
except ImportError:
    orjson = None  # type: ignore


class Request(_Request):
    """
    Custom request class which uses `orjson` if present.
    Otherwise, it falls back to the standard FastAPI request.
    """

    async def body(self) -> bytes:
        if not hasattr(self, "_body"):
            body = await super().body()
            if "gzip" in self.headers.getlist("Content-Encoding"):
                body = gzip_decompress(body)
            self._body = body
        return self._body

    async def json(self) -> Any:
        if orjson is None:
            return await super().json()

        if not hasattr(self, "_json"):
            body = await self.body()
            self._json = orjson.loads(body)

        return self._json



import asyncio

from multiprocessing import Queue
from typing import Awaitable, Callable, Dict, Optional, List

from ..model import MLModel
from ..types import InferenceRequest, InferenceResponse
from ..settings import Settings
from ..env import Environment

from .model import ParallelModel
from .worker import Worker
from .utils import configure_inference_pool, terminate_queue
from .messages import (
    ModelResponseMessage,
    ModelUpdateMessage,
    ModelUpdateType,
)
from .logging import logger
from .dispatcher import Dispatcher


PredictMethod = Callable[[InferenceRequest], Awaitable[InferenceResponse]]
InferencePoolHook = Callable[[Worker], Awaitable[None]]


class InferencePool:
    """
    The InferencePool class represents a pool of workers where we can run
    inference on.

    Under the hood, it's responsible for managing a pool of multiprocessing
    workers, where the model is loaded.
    This approach lets MLServer work around the GIL to make sure that inference
    can occur in parallel across multiple models or instances of a model.
    """

    def __init__(
        self,
        settings: Settings,
        env: Optional[Environment] = None,
        on_worker_stop: List[InferencePoolHook] = [],
    ):
        configure_inference_pool(settings)

        # TODO: Hook on_worker_stop to "worker unexpectedly died" event
        self._on_worker_stop = on_worker_stop
        self._workers: Dict[int, Worker] = {}
        self._settings = settings
        self._responses: Queue[ModelResponseMessage] = Queue()
        for idx in range(self._settings.parallel_workers):
            # TODO: Set callback to restart worker if it goes down (would
            # `worker.join` help with that?)
            worker = Worker(settings, self._responses, env)
            worker.start()
            self._workers[worker.pid] = worker  # type: ignore

        self._dispatcher = Dispatcher(self._workers, self._responses)
        self._dispatcher.start()

    async def load_model(self, model: MLModel) -> MLModel:
        if not self._should_load_model(model):
            # Skip load if model has disabled parallel workers
            return model

        load_message = ModelUpdateMessage(
            update_type=ModelUpdateType.Load,
            model_settings=model.settings,  # type: ignore
        )
        await self._dispatcher.dispatch_update(load_message)

        return ParallelModel(model, self._dispatcher)

    async def reload_model(self, old_model: MLModel, new_model: MLModel) -> MLModel:
        # The model registries within each worker will take care of reloading
        # the model internally
        return await self.load_model(new_model)

    async def unload_model(self, model: MLModel) -> MLModel:
        if not self._should_load_model(model):
            # Skip unload if model has disabled parallel workers
            return model

        unload_message = ModelUpdateMessage(
            update_type=ModelUpdateType.Unload,
            model_settings=model.settings,  # type: ignore
        )
        await self._dispatcher.dispatch_update(unload_message)

        return ParallelModel(model, self._dispatcher)

    def _should_load_model(self, model: MLModel):
        if model.settings.parallel_workers is not None:
            logger.warning(
                "DEPRECATED!! The `parallel_workers` setting at the model-level "
                "has now been deprecated and moved "
                "to the top-level server "
                "settings. "
                "This field will be removed in MLServer 1.2.0. "
                "To access the new field, you can either update the "
                "`settings.json` file, or update the `MLSERVER_PARALLEL_WORKERS` "
                "environment variable. "
                f"The current value of the server-level's `parallel_workers` field is "
                f"'{self._settings.parallel_workers}'."
            )

            # NOTE: This is a remnant from the previous architecture for parallel
            # workers, where each worker had its own pool.
            # For backwards compatibility, we will respect when a model disables
            # parallel inference.
            if model.settings.parallel_workers <= 0:
                return False

        if not self._settings.parallel_workers:
            return False

        return True

    async def close(self):
        logger.info("Waiting for inference pool shutdown")
        await self._close_workers()
        await terminate_queue(self._responses)
        self._responses.close()
        await self._dispatcher.stop()
        logger.info("Inference pool shutdown complete")

    async def _close_workers(self):
        # First close down model updates loop
        for pid, worker in self._workers.items():
            await worker.stop()
            worker.join(self._settings.parallel_workers_timeout)
            if worker.exitcode is None:
                worker.kill()
            await asyncio.gather(
                *[callback(worker) for callback in self._on_worker_stop]
            )

        self._workers.clear()



# Required to deal with annotations including `Queue`
# https://mypy.readthedocs.io/en/latest/common_issues.html#issues-with-code-at-runtime
from __future__ import annotations

import time
import asyncio

from asyncio import Future, Queue, wait_for, Task
from functools import partial
from typing import AsyncIterator, Awaitable, Dict, Optional, Tuple

from ..model import MLModel
from ..types import (
    InferenceRequest,
    InferenceResponse,
)
from ..utils import generate_uuid, schedule_with_callback

from .requests import BatchedRequests

from prometheus_client import Histogram


class AdaptiveBatcher:
    def __init__(self, model: MLModel):
        self._model = model

        self._max_batch_size = model.settings.max_batch_size
        self._max_batch_time = model.settings.max_batch_time

        # Save predict function before it gets decorated
        self._predict_fn = model.predict
        self.__requests: Optional[Queue[Tuple[str, InferenceRequest]]] = None
        self._async_responses: Dict[str, Future[InferenceResponse]] = {}
        self._batching_task = None
        self.batch_request_queue_size = Histogram(
            "batch_request_queue", "counter of request queue batch size"
        )

    async def predict(self, req: InferenceRequest) -> InferenceResponse:
        internal_id, _ = await self._queue_request(req)
        self._start_batcher_if_needed()
        return await self._wait_response(internal_id)

    @property
    def _requests(self) -> Queue[Tuple[str, InferenceRequest]]:
        # NOTE: We need to create Queue within the async request path (and not
        # during __init__!!) to ensure that it shares the same AsyncIO loop.
        if self.__requests is None:
            self.__requests = Queue(maxsize=self._max_batch_size)

        return self.__requests

    async def _queue_request(
        self,
        req: InferenceRequest,
    ) -> Tuple[str, Awaitable[InferenceResponse]]:
        internal_id = generate_uuid()
        self._batch_queue_monitor()
        await self._requests.put((internal_id, req))

        loop = asyncio.get_running_loop()
        async_response = loop.create_future()
        self._async_responses[internal_id] = async_response

        return internal_id, async_response

    def _batch_queue_monitor(self):
        """Monitorize batch queue size"""
        batch_queue_size = self._requests.qsize()
        self.batch_request_queue_size.observe(batch_queue_size)

    async def _wait_response(self, internal_id: str) -> InferenceResponse:
        async_response = self._async_responses[internal_id]

        try:
            response = await async_response
            return response
        finally:
            del self._async_responses[internal_id]

    def _start_batcher_if_needed(self):
        if self._batching_task is not None:
            if not self._batching_task.done():
                # If task hasn't finished yet, let it keep running
                return

        self._batching_task = schedule_with_callback(
            self._batcher(), self._batching_task_callback
        )

    def _batching_task_callback(self, batching_task: Task):
        err = batching_task.exception()
        if err:
            # Clear queue
            self._clear_queue(err)

    def _clear_queue(self, err: BaseException):
        # Cancel all pending async responses
        for async_response in self._async_responses.values():
            async_response.set_exception(err)

        # Empty queue
        for _ in range(self._requests.qsize()):
            self._requests.get_nowait()

    async def _batcher(self):
        async for batched in self._batch_requests():
            # We run prediction as a Task to ensure it gets scheduled
            # immediately.
            # That way, we can process multiple batches concurrently.
            schedule_with_callback(
                self._predict_fn(batched.merged_request),
                partial(self._predict_callback, batched),
            )

    def _predict_callback(self, batched: BatchedRequests, predict_task: Task):
        try:
            batched_response = predict_task.result()
            responses = batched.split_response(batched_response)
            for internal_id, response in responses.items():
                self._async_responses[internal_id].set_result(response)
        except Exception as err:
            for internal_id in batched.inference_requests.keys():
                self._async_responses[internal_id].set_exception(err)

    async def _batch_requests(self) -> AsyncIterator[BatchedRequests]:
        while not self._requests.empty():
            to_batch: Dict[str, InferenceRequest] = {}
            start = time.time()
            timeout = self._max_batch_time

            try:
                while len(to_batch) < self._max_batch_size:
                    internal_id, inference_request = await self._get_request(
                        timeout=timeout
                    )
                    to_batch[internal_id] = inference_request

                    # Update remaining timeout
                    current = time.time()
                    timeout = timeout - (current - start)
            except asyncio.TimeoutError:
                # NOTE: Hit timeout, continue
                pass

            yield BatchedRequests(to_batch)

    async def _get_request(self, timeout: float) -> Tuple[str, InferenceRequest]:
        if not self._requests.empty():
            return await self._requests.get()

        read_op = self._requests.get()
        return await wait_for(read_op, timeout=timeout)



import os
import uuid
import asyncio
import urllib.parse

from asyncio import Task
from typing import Callable, Dict, Optional, List, Type

from .logging import logger
from .types import InferenceRequest, InferenceResponse, Parameters
from .settings import ModelSettings
from .errors import InvalidModelURI


async def get_model_uri(
    settings: ModelSettings, wellknown_filenames: List[str] = []
) -> str:
    if not settings.parameters:
        raise InvalidModelURI(settings.name)

    model_uri = settings.parameters.uri
    if not model_uri:
        raise InvalidModelURI(settings.name)

    model_uri_components = urllib.parse.urlparse(model_uri, scheme="file")
    if model_uri_components.scheme != "file":
        return model_uri

    full_model_path = _to_absolute_path(settings._source, model_uri_components.path)
    if os.path.isfile(full_model_path):
        return full_model_path

    if os.path.isdir(full_model_path):
        # If full_model_path is a folder, search for a well-known model filename
        for fname in wellknown_filenames:
            model_path = os.path.join(full_model_path, fname)
            if os.path.isfile(model_path):
                return model_path

        # If none, return the folder
        return full_model_path

    # Otherwise, the uri is neither a file nor a folder
    raise InvalidModelURI(settings.name, full_model_path)


def _to_absolute_path(source: Optional[str], model_uri: str) -> str:
    if source is None:
        # Treat path as either absolute or relative to the working directory of
        # the MLServer instance
        return model_uri

    parent_folder = os.path.dirname(source)
    unnormalised = os.path.join(parent_folder, model_uri)
    return os.path.normpath(unnormalised)


def get_wrapped_method(f: Callable) -> Callable:
    while hasattr(f, "__wrapped__"):
        f = f.__wrapped__  # type: ignore

    return f


def generate_uuid() -> str:
    return str(uuid.uuid4())


def insert_headers(
    inference_request: InferenceRequest, headers: Dict[str, str]
) -> InferenceRequest:
    # Ensure parameters are present
    if inference_request.parameters is None:
        inference_request.parameters = Parameters()

    parameters = inference_request.parameters

    if parameters.headers is not None:
        # TODO: Raise warning that headers will be replaced and shouldn't be used
        logger.warning(
            f"There are {len(parameters.headers)} entries present in the"
            "`headers` field of the request `parameters` object."
            "The `headers` field of the `parameters` object "
            "SHOULDN'T BE USED directly."
            "These entries will be replaced by the actual headers (REST, Kafka) "
            "or metadata (gRPC) of the incoming request."
        )

    parameters.headers = headers
    return inference_request


def extract_headers(inference_response: InferenceResponse) -> Optional[Dict[str, str]]:
    if inference_response.parameters is None:
        return None

    parameters = inference_response.parameters
    if parameters.headers is None:
        return None

    headers = parameters.headers
    parameters.headers = None
    return headers


def _check_current_event_loop_policy() -> str:
    policy = (
        "uvloop"
        if type(asyncio.get_event_loop_policy()).__module__.startswith("uvloop")
        else "asyncio"
    )
    return policy


def install_uvloop_event_loop():
    if "uvloop" == _check_current_event_loop_policy():
        return

    try:
        import uvloop

        uvloop.install()
    except ImportError:
        # else keep the standard asyncio loop as a fallback
        pass

    policy = _check_current_event_loop_policy()

    logger.debug(f"Using asyncio event-loop policy: {policy}")


def schedule_with_callback(coro, cb) -> Task:
    task = asyncio.create_task(coro)
    task.add_done_callback(cb)
    return task


def get_import_path(klass: Type):
    return f"{klass.__module__}.{klass.__name__}"



import os

from typing import Dict
from setuptools import setup, find_packages

ROOT_PATH = os.path.dirname(__file__)
PKG_NAME = "mlserver"
PKG_PATH = os.path.join(ROOT_PATH, PKG_NAME)


def _load_version() -> str:
    version = ""
    version_path = os.path.join(PKG_PATH, "version.py")
    with open(version_path) as fp:
        version_module: Dict[str, str] = {}
        exec(fp.read(), version_module)
        version = version_module["__version__"]

    return version


def _load_description() -> str:
    readme_path = os.path.join(ROOT_PATH, "README.md")
    with open(readme_path) as fp:
        return fp.read()


env_marker_cpython = (
    "sys_platform != 'win32'"
    " and (sys_platform != 'cygwin'"
    " and platform_python_implementation != 'PyPy')"
)

setup(
    name=PKG_NAME,
    version=_load_version(),
    url="https://github.com/SeldonIO/MLServer.git",
    author="Seldon Technologies Ltd.",
    author_email="hello@seldon.io",
    description="ML server",
    packages=find_packages(exclude=["tests", "tests.*"]),
    install_requires=[
        "click",
        # 0.89.0: https://github.com/tiangolo/fastapi/issues/5861
        "fastapi<=0.89.1, !=0.89.0",
        "python-dotenv",
        "grpcio",
        "importlib-metadata;python_version<'3.8'",
        "numpy",
        "pandas",
        # Force patch for CVE-2022-1941
        # https://github.com/huggingface/optimum/issues/733
        "protobuf == 3.20.3",
        "uvicorn",
        "starlette_exporter",
        "py-grpc-prometheus",
        "uvloop;" + env_marker_cpython,
        "aiokafka",
        "tritonclient[http]>=2.24",
        "aiofiles",
        "orjson",
    ],
    entry_points={"console_scripts": ["mlserver=mlserver.cli:main"]},
    long_description=_load_description(),
    long_description_content_type="text/markdown",
    license="Apache 2.0",
)



import asyncio
import select
import signal

from asyncio import Task, CancelledError
from multiprocessing import Process, Queue
from concurrent.futures import ThreadPoolExecutor
from contextlib import nullcontext
from typing import Optional

from ..registry import MultiModelRegistry
from ..utils import install_uvloop_event_loop, schedule_with_callback
from ..logging import configure_logger
from ..settings import Settings
from ..metrics import configure_metrics
from ..env import Environment

from .messages import (
    ModelRequestMessage,
    ModelUpdateType,
    ModelUpdateMessage,
    ModelResponseMessage,
)
from .utils import terminate_queue, END_OF_QUEUE
from .logging import logger
from .errors import WorkerError

IGNORED_SIGNALS = [signal.SIGINT, signal.SIGTERM, signal.SIGQUIT]


def _noop():
    pass


class Worker(Process):
    def __init__(
        self, settings: Settings, responses: Queue, env: Optional[Environment] = None
    ):
        super().__init__()
        self._settings = settings
        self._responses = responses
        self._requests: Queue[ModelRequestMessage] = Queue()
        self._model_updates: Queue[ModelUpdateMessage] = Queue()
        self._env = env

        self.__executor = None

    @property
    def _executor(self):
        """
        Helper property to initialise a threadpool executor on demand.
        This is required to avoid having to pickle the executor object into a
        separate process.
        """
        if self.__executor is None:
            self.__executor = ThreadPoolExecutor()

        return self.__executor

    def run(self):
        ctx = nullcontext()
        if self._env:
            ctx = self._env

        with ctx:
            install_uvloop_event_loop()
            configure_logger(self._settings)
            configure_metrics(self._settings)
            self._ignore_signals()
            asyncio.run(self.coro_run())

    def _ignore_signals(self):
        """
        Uvloop will try to propagate the main process' signals to the
        underlying workers.
        However, this would just kill off the workers without any cleaning.
        To avoid this, and be able to properly shut them down, we forcefully
        ignore the signals coming from the main parent process.
        """
        loop = asyncio.get_event_loop()

        for sign in IGNORED_SIGNALS:
            # Ensure that signal handlers are a no-op, to let the main process
            # take care of cleaning up workers
            loop.add_signal_handler(sign, _noop)

    def __inner_init__(self):
        """
        Internal __init__ method that needs to run within the worker process.
        """
        self._model_registry = MultiModelRegistry()
        self._active = True

    async def coro_run(self):
        self.__inner_init__()
        loop = asyncio.get_event_loop()

        while self._active:
            readable = await loop.run_in_executor(self._executor, self._select)
            for r in readable:
                if r is self._requests._reader:
                    request = self._requests.get()

                    schedule_with_callback(
                        self._process_request(request), self._handle_response
                    )
                elif r is self._model_updates._reader:
                    model_update = self._model_updates.get()
                    # If the queue gets terminated, detect the "sentinel value"
                    # and stop reading
                    if model_update is END_OF_QUEUE:
                        self._active = False
                        return

                    schedule_with_callback(
                        self._process_model_update(model_update), self._handle_response
                    )

    def _select(self):
        readable, _, _ = select.select(
            [self._requests._reader, self._model_updates._reader],
            [],
            [],
        )

        return readable

    async def _process_request(self, request) -> ModelResponseMessage:
        try:
            model = await self._model_registry.get_model(
                request.model_name, request.model_version
            )

            method = getattr(model, request.method_name)
            return_value = await method(*request.method_args, **request.method_kwargs)
            return ModelResponseMessage(id=request.id, return_value=return_value)
        except (Exception, CancelledError) as e:
            logger.exception(
                f"An error occurred calling method '{request.method_name}' "
                f"from model '{request.model_name}'."
            )
            worker_error = WorkerError(e)
            return ModelResponseMessage(id=request.id, exception=worker_error)

    def _handle_response(self, process_task: Task):
        response_message = process_task.result()
        self._responses.put(response_message)

    async def _process_model_update(
        self, update: ModelUpdateMessage
    ) -> ModelResponseMessage:
        try:
            model_settings = update.model_settings
            if update.update_type == ModelUpdateType.Load:
                await self._model_registry.load(model_settings)
            elif update.update_type == ModelUpdateType.Unload:
                await self._model_registry.unload_version(
                    model_settings.name, model_settings.version
                )
            else:
                logger.warning(
                    "Unknown model update message with type ", update.update_type
                )

            return ModelResponseMessage(id=update.id)
        except (Exception, CancelledError) as e:
            logger.exception(
                "An error occurred processing a model update "
                f"of type '{update.update_type.name}'."
            )
            worker_error = WorkerError(e)
            return ModelResponseMessage(id=update.id, exception=worker_error)

    def send_request(self, request_message: ModelRequestMessage):
        """
        Send an inference request message to the worker.
        Note that this method should be both multiprocess- and thread-safe.
        """
        self._requests.put(request_message)

    def send_update(self, model_update: ModelUpdateMessage):
        """
        Send a model update to the worker.
        Note that this method should be both multiprocess- and thread-safe.
        """
        self._model_updates.put(model_update)

    async def stop(self):
        """
        Close the worker's main loop.
        Note that this method should be both multiprocess- and thread-safe.
        """
        await terminate_queue(self._model_updates)
        self._model_updates.close()
        self._requests.close()
        self._executor.shutdown()



from asyncio import Task

from mlserver.errors import MLServerError
from enum import Enum
from aiokafka import AIOKafkaConsumer, AIOKafkaProducer
from aiokafka.errors import ConsumerStoppedError

from ..handlers import DataPlane
from ..settings import Settings
from ..model import MLModel

from .logging import logger
from .handlers import KafkaHandlers
from .message import KafkaMessage
from ..utils import schedule_with_callback
from typing import Optional


# TODO: Explore implementing custom handler
class KafkaMethodTypes(Enum):
    infer = "infer"


class KafkaServer:
    def __init__(self, settings: Settings, data_plane: DataPlane):
        self._settings = settings
        self._handlers = KafkaHandlers(data_plane)

    def _create_server(self):
        self._consumer = AIOKafkaConsumer(
            self._settings.kafka_topic_input,
            bootstrap_servers=self._settings.kafka_servers,
        )
        self._producer = AIOKafkaProducer(
            bootstrap_servers=self._settings.kafka_servers
        )

    async def add_custom_handlers(self, model: MLModel):
        # TODO: Implement
        pass

    async def delete_custom_handlers(self, model: MLModel):
        # TODO: Implement
        pass

    async def start(self):
        self._create_server()

        await self._consumer.start()
        await self._producer.start()

        logger.info(
            f"Kafka server consuming messages from {self._settings.kafka_servers}."
        )
        try:
            await self._consumer_loop()
        except ConsumerStoppedError:
            logger.info(
                "Stopped consuming messages from topic "
                f"{self._settings.kafka_topic_input}"
            )

    async def _consumer_loop(self):
        logger.info("Reading messages from consumer")
        async for request in self._consumer:
            schedule_with_callback(
                self._process_request(request), self._process_request_cb
            )

    def _process_request_cb(self, process_request_task: Task):
        try:
            process_request_task.result()
        except MLServerError as err:
            logger.exception(f"ERROR {err.status_code} - {str(err)}")
        except Exception as err:
            logger.exception(f"ERROR 500 - {str(err)}")

    async def _process_request(self, request_record):
        kafka_request = KafkaMessage.from_kafka_record(request_record)

        # TODO: Define headers as cloudevent headers
        # TODO: Define standard "method header" and decide values
        # Default if not set is assume inference request
        request_method = kafka_request.headers.get(
            "mlserver-method", KafkaMethodTypes.infer
        )

        # TODO: Move logic to KafkaRouter (Also to implement cutom handlers modularly)
        if request_method != KafkaMethodTypes.infer:
            raise MLServerError(f"Invalid request method: {request_method}")

        kafka_response = await self._handlers.infer(kafka_request)

        await self._producer.send_and_wait(
            self._settings.kafka_topic_output,
            key=kafka_response.encoded_key,
            value=kafka_response.encoded_value,
            headers=kafka_response.encoded_headers,
        )
        logger.info(f"Processed message of type '{request_method}'")

    async def stop(self, sig: Optional[int] = None):
        logger.info("Waiting for Kafka server shutdown")
        await self._consumer.stop()
        await self._producer.stop()
        logger.info("Kafka server shutdown complete")



import subprocess
import os

from tempfile import TemporaryDirectory

from .. import __version__
from ..logging import logger

from .constants import (
    DockerfileName,
    DockerfileTemplate,
    DockerignoreName,
    Dockerignore,
)


def generate_dockerfile() -> str:
    return DockerfileTemplate.format(version=__version__)


def write_dockerfile(
    folder: str, dockerfile: str, include_dockerignore: bool = True
) -> str:
    dockerfile_path = os.path.join(folder, DockerfileName)
    with open(dockerfile_path, "w") as dockerfile_handler:
        logger.info(f"Writing Dockerfile in {dockerfile_path}")
        dockerfile_handler.write(dockerfile)

    if include_dockerignore:
        # Point to our own .dockerignore
        # https://docs.docker.com/engine/reference/commandline/build/#use-a-dockerignore-file
        dockerignore_path = dockerfile_path + DockerignoreName
        with open(dockerignore_path, "w") as dockerignore_handler:
            logger.info(f"Writing .dockerignore in {dockerignore_path}")
            dockerignore_handler.write(Dockerignore)

    return dockerfile_path


def build_image(
    folder: str, dockerfile: str, image_tag: str, no_cache: bool = False
) -> str:
    logger.info(f"Building Docker image with tag {image_tag}")
    _docker_command_prefix = "docker build --rm "
    with TemporaryDirectory() as tmp_dir:
        dockerfile_path = write_dockerfile(tmp_dir, dockerfile)
        _docker_command_suffix = f"{folder} -f {dockerfile_path} -t {image_tag}"
        if no_cache:
            build_cmd = _docker_command_prefix + "--no-cache " + _docker_command_suffix
        else:
            build_cmd = _docker_command_prefix + _docker_command_suffix
        build_env = os.environ.copy()
        build_env["DOCKER_BUILDKIT"] = "1"
        subprocess.run(build_cmd, check=True, shell=True, env=build_env)

    return image_tag



import subprocess
from ..logging import logger


def init_cookiecutter_project(template: str):
    rc = subprocess.call(["which", "cookiecutter"])
    if rc == 0:
        cmd = f"cookiecutter {template}"
        subprocess.run(cmd, check=True, shell=True)
    else:
        logger.error(
            "The cookiecutter command is not found. \n\n"
            "Please install with 'pip install cookiecutter' and retry"
        )



import sys
import os
import json
import importlib

from typing import Any, Dict, List, Optional, Union
from pydantic import BaseSettings, PyObject, Extra, Field
from contextlib import contextmanager

from .version import __version__
from .types import MetadataTensor

ENV_FILE_SETTINGS = ".env"
ENV_PREFIX_SETTINGS = "MLSERVER_"
ENV_PREFIX_MODEL_SETTINGS = "MLSERVER_MODEL_"

DEFAULT_PARALLEL_WORKERS = 1


@contextmanager
def _extra_sys_path(extra_path: str):
    sys.path.insert(0, extra_path)

    yield

    sys.path.remove(extra_path)


def _reload_module(obj: dict):
    import_path = obj.get("implementation", None)
    if not import_path:
        return

    module_path, _, _ = import_path.rpartition(".")
    module = importlib.import_module(module_path)
    importlib.reload(module)


class CORSSettings(BaseSettings):
    class Config:
        env_file = ENV_FILE_SETTINGS
        env_prefix = ENV_PREFIX_SETTINGS

    allow_origins: Optional[List[str]] = []
    """
    A list of origins that should be permitted to make
    cross-origin requests. E.g. ['https://example.org', 'https://www.example.org'].
    You can use ['*'] to allow any origin
    """

    allow_origin_regex: Optional[str] = None
    """
    A regex string to match against origins that
    should be permitted to make cross-origin requests.
    e.g. 'https:\\/\\/.*\\.example\\.org'
    """

    allow_credentials: Optional[bool] = False
    """Indicate that cookies should be supported for cross-origin requests"""

    allow_methods: Optional[List[str]] = ["GET"]
    """A list of HTTP methods that should be allowed for cross-origin requests"""

    allow_headers: Optional[List[str]] = []
    """A list of HTTP request headers that should be supported for
    cross-origin requests"""

    expose_headers: Optional[List[str]] = []
    """Indicate any response headers that should be made accessible to the browser"""

    max_age: Optional[int] = 600
    """Sets a maximum time in seconds for browsers to cache CORS responses"""


class Settings(BaseSettings):
    class Config:
        env_file = ENV_FILE_SETTINGS
        env_prefix = ENV_PREFIX_SETTINGS

    debug: bool = True

    parallel_workers: int = DEFAULT_PARALLEL_WORKERS
    """When parallel inference is enabled, number of workers to run inference
    across."""

    parallel_workers_timeout: int = 5
    """Grace timeout to wait until the workers shut down when stopping MLServer."""

    # Custom model repository class implementation
    model_repository_implementation: Optional[PyObject] = None
    """*Python path* to the inference runtime to model repository (e.g.
    ``mlserver.repository.repository.SchemalessModelRepository``)."""

    # Model repository settings
    model_repository_root: str = "."
    """Root of the model repository, where we will search for models."""

    # Model Repository parameters are meant to be set directly by the MLServer runtime.
    model_repository_implementation_args: dict = {}
    """Extra parameters for model repository."""

    load_models_at_startup: bool = True
    """Flag to load all available models automatically at startup."""

    # Server metadata
    server_name: str = "mlserver"
    """Name of the server."""

    server_version: str = __version__
    """Version of the server."""

    extensions: List[str] = []
    """Server extensions loaded."""

    # HTTP Server settings
    host: str = "0.0.0.0"
    """Host where to listen for connections."""

    http_port: int = 8080
    """Port where to listen for HTTP / REST connections."""

    root_path: str = ""
    """Set the ASGI root_path for applications submounted below a given URL path."""

    grpc_port: int = 8081
    """Port where to listen for gRPC connections."""

    grpc_max_message_length: Optional[int] = None
    """Maximum length (i.e. size) of gRPC payloads."""

    # CORS settings
    cors_settings: Optional[CORSSettings] = None

    # Metrics settings
    metrics_endpoint: Optional[str] = "/metrics"
    """
    Endpoint used to expose Prometheus metrics. Alternatively, can be set to
    `None` to disable it.
    """

    metrics_port: int = 8082
    """
    Port used to expose metrics endpoint.
    """

    metrics_rest_server_prefix: str = "rest_server"
    """
    Metrics rest server string prefix to be exported.
    """

    metrics_dir: str = os.getcwd()
    """
    Directory used to share metrics across parallel workers.
    Equivalent to the `PROMETHEUS_MULTIPROC_DIR` env var in
    `prometheus-client`.
    Note that this won't be used if the `parallel_workers` flag is disabled.
    """

    # Logging settings
    logging_settings: Optional[Union[str, Dict]] = None
    """Path to logging config file or dictionary configuration."""

    # Kakfa Server settings
    kafka_enabled: bool = False
    kafka_servers: str = "localhost:9092"
    kafka_topic_input: str = "mlserver-input"
    kafka_topic_output: str = "mlserver-output"

    # Custom server settings
    _custom_rest_server_settings: Optional[dict] = None
    _custom_metrics_server_settings: Optional[dict] = None
    _custom_grpc_server_settings: Optional[dict] = None


class ModelParameters(BaseSettings):
    """
    Parameters that apply only to a particular instance of a model.
    This can include things like model weights, or arbitrary ``extra``
    parameters particular to the underlying inference runtime.
    The main difference with respect to ``ModelSettings`` is that parameters
    can change on each instance (e.g. each version) of the model.
    """

    class Config:
        extra = Extra.allow
        env_file = ENV_FILE_SETTINGS
        env_prefix = ENV_PREFIX_MODEL_SETTINGS

    uri: Optional[str] = None
    """
    URI where the model artifacts can be found.
    This path must be either absolute or relative to where MLServer is running.
    """

    version: Optional[str] = None
    """Version of the model."""

    format: Optional[str] = None
    """Format of the model (only available on certain runtimes)."""

    content_type: Optional[str] = None
    """Default content type to use for requests and responses."""

    extra: Optional[dict] = {}
    """Arbitrary settings, dependent on the inference runtime
    implementation."""


class ModelSettings(BaseSettings):
    class Config:
        env_file = ENV_FILE_SETTINGS
        env_prefix = ENV_PREFIX_MODEL_SETTINGS
        underscore_attrs_are_private = True

    # Source points to the file where model settings were loaded from
    _source: Optional[str] = None

    @classmethod
    def parse_file(cls, path: str) -> "ModelSettings":  # type: ignore
        with open(path, "r") as f:
            obj = json.load(f)
            obj["_source"] = path
            return cls.parse_obj(obj)

    @classmethod
    def parse_obj(cls, obj: Any) -> "ModelSettings":
        source = obj.pop("_source", None)
        if not source:
            return super().parse_obj(obj)

        model_folder = os.path.dirname(source)
        with _extra_sys_path(model_folder):
            _reload_module(obj)
            model_settings = super().parse_obj(obj)
            model_settings._source = source
            return model_settings

    @property
    def version(self) -> Optional[str]:
        params = self.parameters
        if params is not None:
            return params.version
        return None

    name: str = ""
    """Name of the model."""

    # Model metadata
    platform: str = ""
    """Framework used to train and serialise the model (e.g. sklearn)."""

    versions: List[str] = []
    """Versions of dependencies used to train the model (e.g.
    sklearn/0.20.1)."""

    inputs: List[MetadataTensor] = []
    """Metadata about the inputs accepted by the model."""

    outputs: List[MetadataTensor] = []
    """Metadata about the outputs returned by the model."""

    # Parallel settings
    parallel_workers: Optional[int] = Field(
        None,
        deprecated=True,
        description=(
            "Use the `parallel_workers` field the server wide settings instead."
        ),
    )

    warm_workers: bool = Field(
        False,
        deprecated=True,
        description="Inference workers will now always be `warmed up` at start time.",
    )

    # Adaptive Batching settings (disabled by default)
    max_batch_size: int = 0
    """When adaptive batching is enabled, maximum number of requests to group
    together in a single batch."""

    max_batch_time: float = 0.0
    """When adaptive batching is enabled, maximum amount of time (in seconds)
    to wait for enough requests to build a full batch."""

    # Custom model class implementation
    implementation: PyObject
    """*Python path* to the inference runtime to use to serve this model (e.g.
    ``mlserver_sklearn.SKLearnModel``)."""

    # Model parameters are meant to be set directly by the MLServer runtime.
    # However, it's also possible to override them manually.
    parameters: Optional[ModelParameters] = None
    """Extra parameters for each instance of this model."""
